import pytest
import numpy as np
from libertem.common import Slice, Shape, SliceUsageError


def test_subslices_simple():
    top_slice = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((4, 4, 4, 4), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 2, 4, 4))) == [
        Slice(origin=(0, 0, 0, 0), shape=Shape((2, 2, 4, 4), sig_dims=2)),
        Slice(origin=(0, 2, 0, 0), shape=Shape((2, 2, 4, 4), sig_dims=2)),
        Slice(origin=(2, 0, 0, 0), shape=Shape((2, 2, 4, 4), sig_dims=2)),
        Slice(origin=(2, 2, 0, 0), shape=Shape((2, 2, 4, 4), sig_dims=2)),
    ]


def test_subslices_non_even_division_1():
    top_slice = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((5, 1, 1, 1), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 1, 1, 1))) == [
        Slice(origin=(0, 0, 0, 0), shape=Shape((2, 1, 1, 1), sig_dims=2)),
        Slice(origin=(2, 0, 0, 0), shape=Shape((2, 1, 1, 1), sig_dims=2)),
        Slice(origin=(4, 0, 0, 0), shape=Shape((1, 1, 1, 1), sig_dims=2)),
    ]


def test_subslices_non_even_division_2():
    top_slice = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((3, 1, 1, 1), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 1, 1, 1))) == [
        Slice(origin=(0, 0, 0, 0), shape=Shape((2, 1, 1, 1), sig_dims=2)),
        Slice(origin=(2, 0, 0, 0), shape=Shape((1, 1, 1, 1), sig_dims=2)),
    ]


def test_subslices_non_even_division_3():
    top_slice = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((3, 3, 1, 1), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 2, 1, 1))) == [
        Slice(origin=(0, 0, 0, 0), shape=Shape((2, 2, 1, 1), sig_dims=2)),
        Slice(origin=(0, 2, 0, 0), shape=Shape((2, 1, 1, 1), sig_dims=2)),
        Slice(origin=(2, 0, 0, 0), shape=Shape((1, 2, 1, 1), sig_dims=2)),
        Slice(origin=(2, 2, 0, 0), shape=Shape((1, 1, 1, 1), sig_dims=2)),
    ]


def test_subslices_non_even_division_4():
    top_slice = Slice(
        origin=(1, 0, 0, 0),
        shape=Shape((1, 10, 3838, 3710), sig_dims=2),
    )
    list(top_slice.subslices(shape=(1, 2, 128, 128)))


def test_subslices_non_even_division_with_origin_1():
    top_slice = Slice(
        origin=(0, 3, 0, 0),
        shape=Shape((3, 3, 1, 1), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 2, 1, 1))) == [
        Slice(origin=(0, 3, 0, 0), shape=Shape((2, 2, 1, 1), sig_dims=2)),
        Slice(origin=(0, 5, 0, 0), shape=Shape((2, 1, 1, 1), sig_dims=2)),
        Slice(origin=(2, 3, 0, 0), shape=Shape((1, 2, 1, 1), sig_dims=2)),
        Slice(origin=(2, 5, 0, 0), shape=Shape((1, 1, 1, 1), sig_dims=2)),
    ]


def test_subslices_non_even_division_with_origin_2():
    top_slice = Slice(
        origin=(0, 3, 0, 0),
        shape=Shape((3, 3, 3, 3), sig_dims=2),
    )
    assert list(top_slice.subslices(shape=(2, 2, 2, 2))) == [
        Slice(origin=(0, 3, 0, 0), shape=Shape((2, 2, 2, 2), sig_dims=2)),
        Slice(origin=(0, 3, 0, 2), shape=Shape((2, 2, 2, 1), sig_dims=2)),
        Slice(origin=(0, 3, 2, 0), shape=Shape((2, 2, 1, 2), sig_dims=2)),
        Slice(origin=(0, 3, 2, 2), shape=Shape((2, 2, 1, 1), sig_dims=2)),

        Slice(origin=(0, 5, 0, 0), shape=Shape((2, 1, 2, 2), sig_dims=2)),
        Slice(origin=(0, 5, 0, 2), shape=Shape((2, 1, 2, 1), sig_dims=2)),
        Slice(origin=(0, 5, 2, 0), shape=Shape((2, 1, 1, 2), sig_dims=2)),
        Slice(origin=(0, 5, 2, 2), shape=Shape((2, 1, 1, 1), sig_dims=2)),

        Slice(origin=(2, 3, 0, 0), shape=Shape((1, 2, 2, 2), sig_dims=2)),
        Slice(origin=(2, 3, 0, 2), shape=Shape((1, 2, 2, 1), sig_dims=2)),
        Slice(origin=(2, 3, 2, 0), shape=Shape((1, 2, 1, 2), sig_dims=2)),
        Slice(origin=(2, 3, 2, 2), shape=Shape((1, 2, 1, 1), sig_dims=2)),

        Slice(origin=(2, 5, 0, 0), shape=Shape((1, 1, 2, 2), sig_dims=2)),
        Slice(origin=(2, 5, 0, 2), shape=Shape((1, 1, 2, 1), sig_dims=2)),
        Slice(origin=(2, 5, 2, 0), shape=Shape((1, 1, 1, 2), sig_dims=2)),
        Slice(origin=(2, 5, 2, 2), shape=Shape((1, 1, 1, 1), sig_dims=2)),
    ]


def test_get_slice_1():
    slice_ = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((4, 4, 4, 4), sig_dims=2),
    )
    assert slice_.get() == (
        slice(0, 4),
        slice(0, 4),
        slice(0, 4),
        slice(0, 4),
    )


def test_get_slice_2():
    slice_ = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((1, 1, 2, 2), sig_dims=2),
    )
    data = np.arange(4 * 4 * 4 * 4).reshape(4, 4, 4, 4)
    assert slice_.get(data).shape == tuple(slice_.shape)
    assert np.all(slice_.get(data) == np.array([[[
        [85, 86],
        [89, 90],
    ]]]))


def test_get_slice_stack_signal_only():
    slice_ = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2)
    )

    data = np.arange(4 * 4 * 4 * 4).reshape(4, 4, 4, 4)
    assert slice_.get(data, sig_only=True).shape[2:4] == tuple(slice_.shape.sig)
    assert np.all(slice_.get(data, sig_only=True) == data[..., 0:1, 0:1])


def test_get_slice_stack_nav_only():
    slice_ = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2)
    )

    data = np.arange(4 * 4 * 4 * 4).reshape(4, 4, 4, 4)
    assert slice_.get(data, nav_only=True).shape[0:2] == tuple(slice_.shape.nav)
    assert np.all(slice_.get(data, nav_only=True) == data[0:1, 0:1])


def test_slice_intersect_0():
    s1 = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((2, 2, 2, 2), sig_dims=2),
    )
    s2 = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2),
    )
    assert s1.intersection_with(s2) == s2


def test_slice_intersect_1():
    s1 = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((2, 2, 2, 2), sig_dims=2),
    )
    s2 = Slice(
        origin=(3, 3, 3, 3),
        shape=Shape((1, 1, 1, 1), sig_dims=2),
    )
    res = s1.intersection_with(s2)
    assert res == Slice(
        origin=(3, 3, 3, 3),
        shape=Shape((0, 0, 0, 0), sig_dims=2),
    )
    assert res.is_null()


def test_slice_intersect_2():
    s1 = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((2, 2, 2, 2), sig_dims=2),
    )
    s2 = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2),
    )
    res = s1.intersection_with(s2)
    assert res == Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((0, 0, 0, 0), sig_dims=2),
    )
    assert res.is_null()


def test_slice_intersect_3():
    s1 = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((2, 2, 2, 2), sig_dims=2)
    )
    s2 = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((4, 4, 4, 4), sig_dims=2)
    )
    res = s1.intersection_with(s2)
    assert res == s1


def test_shift_1():
    s1 = Slice(
        origin=(1, 1, 0, 0),
        shape=Shape((1, 1, 2, 2), sig_dims=2)
    )

    s2 = Slice(
        origin=(1, 1, 0, 0),
        shape=Shape((1, 1, 4, 4), sig_dims=2)
    )

    shifted = s1.shift(s2)

    assert shifted.origin == (0, 0, 0, 0)


def test_shift_2():
    s1 = Slice(
        origin=(2, 2, 0, 0),
        shape=Shape((1, 1, 2, 2), sig_dims=2)
    )

    s2 = Slice(
        origin=(1, 1, 0, 0),
        shape=Shape((1, 1, 4, 4), sig_dims=2)
    )

    shifted = s1.shift(s2)
    assert shifted.origin == (1, 1, 0, 0)


def test_get_signal_only():
    s = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2)
    )

    assert s.get(sig_only=True) == (
        slice(0, 1),
        slice(0, 1),
    )


def test_get():
    s = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2)
    )

    assert s.get() == (
        slice(0, 1),
        slice(0, 1),
        slice(0, 1),
        slice(0, 1),
    )


def test_flatten_nav():
    s = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((1, 1, 1, 1), sig_dims=2)
    )
    sflat = Slice(
        origin=(0, 0, 0),
        shape=Shape((1, 1, 1), sig_dims=2)
    )
    assert s.flatten_nav((1, 1, 1, 1)) == sflat


def test_flatten_nav_2():
    s = Slice(
        origin=(0, 0, 0, 0),
        shape=Shape((2, 16, 16, 16), sig_dims=2)
    )
    sflat = Slice(
        origin=(0, 0, 0),
        shape=Shape((32, 16, 16), sig_dims=2)
    )
    assert s.flatten_nav((16, 16, 16, 16)) == sflat


def test_slice_to_sig():
    s = Slice(
        origin=(0, 0, 8, 8),
        shape=Shape((2, 16, 16, 16), sig_dims=2)
    )
    assert s.sig.origin == (8, 8)
    assert s.sig.shape.to_tuple() == (16, 16)
    assert s.sig.shape.sig.dims == 2


def test_slice_to_nav():
    s = Slice(
        origin=(3, 3, 8, 8),
        shape=Shape((2, 16, 16, 16), sig_dims=2)
    )
    assert s.nav.origin == (3, 3)
    assert s.nav.shape.to_tuple() == (2, 16)
    assert s.nav.shape.sig.dims == 0


def test_from_shape():
    s = Slice.from_shape(
        (1, 16, 16),
        sig_dims=2
    )
    assert s == Slice(
        origin=(0, 0, 0),
        shape=Shape((1, 16, 16), sig_dims=2),
    )


def test_shift_offset_consistency():
    offset1 = (0, -3, 5)
    offset2 = (1, 7, -19)
    total_offset = tuple(o1 + o2 for (o1, o2) in zip(offset1, offset2))
    s1 = Slice.from_shape(
        (1, 16, 16),
        sig_dims=2
    )
    s2 = s1.shift_by(offset1)
    s3 = s2.shift_by(offset2)
    assert s3.origin == total_offset
    assert s3.shape == s1.shape


def test_slice_raises_not_shape():
    with pytest.raises(SliceUsageError):
        Slice(
            origin=(1, 1, 1, 1),
            shape=(2, 2, 2, 2),
        )


def test_slice_raises_mismatching():
    with pytest.raises(SliceUsageError):
        Slice(
            origin=(1, 1, 1),
            shape=Shape((2, 2, 2, 2), sig_dims=2),
        )


def test_intersections_raises_mismatching():
    s1 = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((2, 2, 2, 2), sig_dims=2),
    )
    s2 = Slice(
        origin=(1, 1, 1),
        shape=Shape((2, 2, 2), sig_dims=2),
    )
    s3 = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((2, 2, 2, 2), sig_dims=1),
    )
    with pytest.raises(SliceUsageError):
        s1.intersection_with(s2)
    with pytest.raises(SliceUsageError):
        s1.intersection_with(s3)


def test_shifts_raises_mismatching():
    s1 = Slice(
        origin=(1, 1, 1, 1),
        shape=Shape((2, 2, 2, 2), sig_dims=2),
    )
    s2 = Slice(
        origin=(1, 1, 1),
        shape=Shape((2, 2, 2), sig_dims=2),
    )
    with pytest.raises(SliceUsageError):
        s1.shift(s2)
    with pytest.raises(SliceUsageError):
        s1.shift_by(s2.origin)



from typing import Any
from collections.abc import Iterable
import contextlib

import cloudpickle
import psutil
import uuid

from .base import BaseJobExecutor
from libertem.common.executor import (
    Environment, SimpleWorkerQueue, TaskCommHandler, TaskProtocol, WorkerContext, WorkerQueue
)
from libertem.common.scheduler import Worker, WorkerSet
from libertem.common.backend import get_use_cuda


class InlineWorkerContext(WorkerContext):
    def __init__(self, queue: SimpleWorkerQueue, msg_queue: SimpleWorkerQueue):
        self._queue = queue
        self._msg_queue = msg_queue

    def get_worker_queue(self) -> WorkerQueue:
        return self._queue

    def signal(self, ident: str, topic: str, msg_dict: dict[str, Any]):
        if 'ident' in msg_dict:
            raise ValueError('ident is a reserved name')
        msg_dict.update({'ident': ident})
        self._msg_queue.put((topic, msg_dict))


class InlineJobExecutor(BaseJobExecutor):
    """
    Naive JobExecutor that just iterates over partitions and processes them one after another

    Parameters
    ----------
    debug : bool
        Set this to enable additional serializability checks

    inline_threads : Optional[int]
        How many fine grained threads should be allowed? Leaving this `None` will
        allow one thread per CPU core
    """
    def __init__(self, debug=False, inline_threads=None, *args, **kwargs):
        # Only import if actually instantiated, i.e. will likely be used
        import libertem.preload  # noqa: 401
        self._debug = debug
        self._inline_threads = inline_threads
        self._scattered = {}

    @contextlib.contextmanager
    def scatter(self, obj):
        obj_id = uuid.uuid4()
        self._scattered[obj_id] = obj
        try:
            yield obj_id
        finally:
            del self._scattered[obj_id]

    def scatter_update(self, handle, obj):
        self._scattered[handle] = obj

    def scatter_update_patch(self, handle, patch):
        self._scattered[handle].patch(patch)

    def run_tasks(
        self,
        tasks: Iterable[TaskProtocol],
        params_handle: Any,
        cancel_id: Any,
        task_comm_handler: TaskCommHandler,
    ):
        worker_queue = SimpleWorkerQueue()
        msg_queue = SimpleWorkerQueue()
        task_comm_handler.start()
        threads = self._inline_threads
        if threads is None:
            threads = psutil.cpu_count(logical=False)
        worker_context = InlineWorkerContext(queue=worker_queue,
                                             msg_queue=msg_queue)
        env = Environment(
            threads_per_worker=threads,
            threaded_executor=False,
            worker_context=worker_context,
        )
        with task_comm_handler.monitor(msg_queue):
            for task in tasks:
                if self._debug:
                    cloudpickle.loads(cloudpickle.dumps(task))
                task_comm_handler.handle_task(task, worker_queue)
                result = task(env=env, params=self._scattered[params_handle])
                if self._debug:
                    cloudpickle.loads(cloudpickle.dumps(result))
                yield result, task

        task_comm_handler.done()

    def run_function(self, fn, *args, **kwargs):
        if self._debug:
            cloudpickle.loads(cloudpickle.dumps((fn, args, kwargs)))
        result = fn(*args, **kwargs)
        if self._debug:
            cloudpickle.loads(cloudpickle.dumps(result))
        return result

    def map(self, fn, iterable):
        return [fn(item)
                for item in iterable]

    def run_each_host(self, fn, *args, **kwargs):
        if self._debug:
            cloudpickle.loads(cloudpickle.dumps((fn, args, kwargs)))
        return {"localhost": fn(*args, **kwargs)}

    def run_each_worker(self, fn, *args, **kwargs):
        return {"inline": fn(*args, **kwargs)}

    def get_available_workers(self):
        resources = {"compute": 1, "CPU": 1}
        if get_use_cuda() is not None:
            resources["CUDA"] = 1

        return WorkerSet([
            Worker(name='inline', host='localhost', resources=resources, nthreads=1)
        ])



import numpy as np

from libertem.udf.base import UDFRunner
from libertem.executor.inline import InlineJobExecutor
from libertem.udf.sum import SumUDF
from libertem.udf.sumsigudf import SumSigUDF
from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random


def test_simple_multi_udf_run():
    data = _mk_random(size=(32, 1860, 2048))
    dataset = MemoryDataSet(
        data=data,
        num_partitions=1,
        sig_dims=2,
        base_shape=(1, 930, 16),
        force_need_decode=True,
    )

    executor = InlineJobExecutor()
    udfs = [
        SumSigUDF(),
        SumUDF(),
    ]
    res = UDFRunner(udfs=udfs).run_for_dataset(
        dataset=dataset,
        executor=executor,
    )
    sumsigres, sumres = res.buffers
    print(sumsigres, sumres)
    assert np.allclose(
        sumres['intensity'],
        np.sum(data, axis=0)
    )
    assert np.allclose(
        sumsigres['intensity'],
        np.sum(data, axis=(1, 2))
    )



import sys
import msvcrt
import ctypes
from ctypes import windll, wintypes, byref
import logging
import warnings


logger = logging.getLogger(__name__)


def get_owner_name_dummy(full_path, stat):
    return "<Unknown>"


try:
    import win32security
    import pywintypes

    def get_owner_name(full_path, stat):
        try:
            s = win32security.GetFileSecurity(full_path, win32security.OWNER_SECURITY_INFORMATION)
            sid = s.GetSecurityDescriptorOwner()
            (name, domain, t) = win32security.LookupAccountSid(None, sid)
            return f"{domain}\\{name}"
        except pywintypes.error as e:
            raise OSError(e)

except ModuleNotFoundError:
    warnings.warn("Install package pywin32 to enable detecting file owners on Windows.")

    get_owner_name = get_owner_name_dummy

except ImportError as e:
    warnings.warn(
        f"Importing modules from pywin32 failed. {e} "
        "Possibly the post installation script of pywin32 has to be run. More information: "
        "https://github.com/mhammond/pywin32#installing-via-pip"
    )

    get_owner_name = get_owner_name_dummy


ENABLE_QUICK_EDIT = 0x0040
ENABLE_EXTENDED_FLAGS = 0x0080


def get_console_mode(stream=sys.stdin):
    file_handle = msvcrt.get_osfhandle(stream.fileno())
    getConsoleMode = windll.kernel32.GetConsoleMode
    getConsoleMode.argtypes, getConsoleMode.restype = ([wintypes.HANDLE, wintypes.LPDWORD],
                                                       wintypes.BOOL)
    mode = wintypes.DWORD(0)
    if getConsoleMode(file_handle, byref(mode)):
        return mode.value
    else:
        err = ctypes.get_last_error()
        raise ctypes.WinError(err)


def set_console_mode(mode, stream=sys.stdin):
    file_handle = msvcrt.get_osfhandle(stream.fileno())
    setConsoleMode = windll.kernel32.SetConsoleMode
    setConsoleMode.argtypes, setConsoleMode.restype = ([wintypes.HANDLE, wintypes.DWORD],
                                                       wintypes.BOOL)
    if setConsoleMode(file_handle, mode):
        return
    else:
        err = ctypes.get_last_error()
        raise ctypes.WinError(err)


def disable_quickedit():
    try:
        mode = get_console_mode()
        mode &= ~ENABLE_QUICK_EDIT
        mode |= ENABLE_EXTENDED_FLAGS
        set_console_mode(mode)
    except OSError:
        logger.info("Quick Edit couldn't be disabled. Console probably doesn't support Quick Edit.")


if __name__ == "__main__":
    disable_quickedit()



import os
import platform
import tempfile

import numpy as np


class WriteHandle:
    """
    Parameters:
    -----------
    path : str
        Full path to the file to be written (should not exist, but the directory should)

    tmp_base_path : str
        Path to a directory where temporary files should be written
        (must be on the same file system as `path`, so maybe)

    part_slice : Slice
        Slice of the object (i.e. Partition) we are writing, to convert
        from global tile coordinates to local slices.

    dtype : numpy dtype
        For which dtype should the file be opened
    """
    def __init__(self, path, tmp_base_path, part_slice, dtype):
        # TODO: support for direct I/O
        # (useful if we have a very high write rate, otherwise not so much; very high being
        # multiple GiB/s)
        self._path = path
        self._tmp_base_path = tmp_base_path
        self._slice = part_slice
        self._dtype = dtype
        self._dest = None
        self._tmp_file = None
        self._aborted = False

    def write_tile(self, tile):
        """
        Write a single `DataTile`
        """
        assert self._tmp_file is not None
        dest_slice = tile.tile_slice.shift(self._slice)
        self._dest[dest_slice.get()] = tile.data

    def write_tiles(self, tiles):
        """
        Write all `tiles`, while yielding each tile for further processing
        """
        for tile in tiles:
            self.write_tile(tile)
            yield tile

    def abort(self):
        self._cleanup()
        self._aborted = True

    def __enter__(self):
        self._open_tmp()
        shape = tuple(self._slice.shape)
        self._dest = np.memmap(self._tmp_file.name, dtype=self._dtype, mode='write', shape=shape)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None or self._aborted:
            # we have an exception, or abort() was called
            self._cleanup()
            return
        self._finalize()

    def _open_tmp(self):
        """
        Open a temporary file for writing. We pass delete=False, because, on success, we move the
        file to its new name, and don't want it to be deleted. Only in case of errors should
        it be removed (done in _cleanup).

        Filename can be accessed as `self._tmp_file.name`
        """
        assert self._tmp_file is None
        prefix = os.path.basename(".tmp-%s" % self._path)
        self._tmp_file = tempfile.NamedTemporaryFile(
            prefix=prefix,
            dir=self._tmp_base_path,
            delete=False
        )

    def _cleanup(self):
        """
        In case of errors, remove the temporary file.
        """
        self._dest = None
        if self._tmp_file is not None:
            self._tmp_file.close()
            os.unlink(self._tmp_file.name)
            self._tmp_file = None

    def _finalize(self):
        """
        Called in case of success.

        What we need to do here:
         + call msync(2) w/ MS_SYNC to flush changes to filesystem
           (done by numpy when caling `flush` on the memmap object)
         + atomically move partition to final filename
         + call fsync on the destination directory
        """
        self._dest.flush()  # msync

        # to make Windows™ happy:
        self._tmp_file.close()

        # FIXME temporary workaround, see if fixed upstream:
        # https://github.com/numpy/numpy/issues/13510
        mm = self._dest._mmap
        if mm is not None:
            mm.close()

        os.rename(self._tmp_file.name, self._path)
        dest_dir = os.path.dirname(self._path)
        self.sync_dir(dest_dir)
        self._tmp_file = None
        self._dest = None

    def sync_dir(self, path):
        # noop on windows, as:
        # "On Windows… err, there is no clear answer. You can not call FlushFileBuffers()
        # on a directory handle as far as I can see."
        # (from http://blog.httrack.com/blog/2013/11/15/everything-you-always-wanted-to-know-about-fsync/)  # NOQA
        if platform.system() == "Windows":
            return
        fd = os.open(path, os.O_RDONLY | os.O_DIRECTORY)
        try:
            os.fsync(fd)
        finally:
            os.close(fd)



import warnings
from typing import Optional, TYPE_CHECKING

import numpy as np
from sparseconverter import ArrayBackend, for_backend

from libertem.common import Slice, Shape
from libertem.common.math import count_nonzero
from libertem.io.corrections import CorrectionSet
from .tiling import DataTile
from .tiling_scheme import TilingScheme
from .meta import DataSetMeta
from .fileset import FileSet
from . import IOBackend
from .decode import Decoder
from .roi import roi_for_partition


if TYPE_CHECKING:
    from libertem.common.executor import WorkerContext


class WritablePartition:
    def get_write_handle(self):
        raise NotImplementedError()

    def delete(self):
        raise NotImplementedError()


class Partition:
    """
    Parameters
    ----------
    meta
        The `DataSet`'s `DataSetMeta` instance

    partition_slice
        The partition slice in non-flattened form

    fileset
        The files that are part of this partition (the FileSet may also contain files
        from the dataset which are not part of this partition, but that may harm performance)

    io_backend
        The I/O backend to use for accessing this partition

    decoder
        The decoder that needs to be used for decoding this partition's data
    """
    def __init__(
        self,
        meta: DataSetMeta,
        partition_slice: Slice,
        io_backend: IOBackend,
        decoder: Optional[Decoder],
    ):
        self.meta = meta
        self.slice = partition_slice
        self._io_backend = io_backend
        self._decoder = decoder
        self._idx: Optional[int] = None
        if partition_slice.shape.nav.dims != 1:
            raise ValueError("nav dims should be flat")

    @classmethod
    def make_slices(cls, shape, num_partitions, sync_offset=0):
        """
        partition a 3D dataset ("list of frames") along the first axis,
        yielding the partition slice, and additionally start and stop frame
        indices for each partition.
        """
        num_frames = shape.nav.size
        if num_partitions > num_frames:
            warnings.warn(
                "dataset contains fewer frames than partitions, "
                "setting num_part to 1 to allow processing (use fewer workers?)",
                RuntimeWarning
            )
            num_partitions = 1
        boundaries = np.linspace(
            0,
            num_frames,
            num=max(2, num_partitions + 1),
            endpoint=True,
            dtype=int,
        )

        # Cast explicitly to tuple[int, ...] to avoid pickle/JSON errors
        boundaries = tuple(map(int, boundaries))

        for (start, stop) in zip(boundaries[:-1], boundaries[1:]):
            part_slice = Slice(
                origin=(start,) + tuple([0] * shape.sig.dims),
                shape=Shape(((stop - start),) + tuple(shape.sig),
                            sig_dims=shape.sig.dims)
            )
            yield part_slice, start + sync_offset, stop + sync_offset

    def set_io_backend(self, backend):
        raise NotImplementedError()

    def validate_tiling_scheme(self, tiling_scheme):
        pass

    def set_corrections(self, corrections: CorrectionSet):
        raise NotImplementedError()

    def set_worker_context(self, worker_context: "WorkerContext"):
        pass

    def get_tiles(self, tiling_scheme, dest_dtype="float32", roi=None,
            array_backend: Optional[ArrayBackend] = None):
        raise NotImplementedError()

    def __repr__(self):
        return "<{}>".format(
            self.__class__.__name__,
        )

    @property
    def dtype(self):
        return self.meta.dtype

    @property
    def shape(self) -> Shape:
        """
        the shape of the partition; dimensionality depends on format
        """
        return self.slice.shape.flatten_nav()

    def get_macrotile(self, dest_dtype="float32", roi=None,
            array_backend: Optional[ArrayBackend] = None):
        '''
        Return a single tile for the entire partition.

        This is useful to support process_partiton() in UDFs and to construct dask arrays
        from datasets.
        '''

        tiling_scheme = TilingScheme.make_for_shape(
            tileshape=self.shape,
            dataset_shape=self.meta.shape,
        )

        try:
            tiles = self.get_tiles(
                tiling_scheme=tiling_scheme,
                dest_dtype=dest_dtype,
                roi=roi,
                array_backend=array_backend,
            )
            tile = next(tiles)
            # NOTE: run the generator to completion, but there must not be any
            # more tiles than the one!
            rest = list(tiles)
            assert len(rest) == 0
            return tile
        except StopIteration:
            tile_slice = Slice(
                origin=(self.slice.origin[0], 0, 0),
                shape=Shape((0,) + tuple(self.slice.shape.sig), sig_dims=2),
            )
            empty = np.zeros(tile_slice.shape, dtype=dest_dtype)
            return DataTile(
                for_backend(empty, array_backend, strict=False),
                tile_slice=tile_slice,
                scheme_idx=0,
            )

    def get_locations(self):
        raise NotImplementedError()

    def get_io_backend(self):
        return None

    def set_idx(self, idx: int):
        self._idx = idx

    def get_ident(self) -> str:
        return f'part-{self._idx}'

    def get_frame_count(self, roi: Optional[np.ndarray] = None) -> int:
        if roi is None:
            return self.shape.nav.size
        else:
            return count_nonzero(roi_for_partition(roi, self))


class BasePartition(Partition):
    """
    Base class with default implementations

    Parameters
    ----------
    meta
        The `DataSet`'s `DataSetMeta` instance

    partition_slice
        The partition slice in non-flattened form

    fileset
        The files that are part of this partition (the FileSet may also contain files
        from the dataset which are not part of this partition, but that may harm performance)

    start_frame
        The index of the first frame of this partition (global coords)

    num_frames
        How many frames this partition should contain

    io_backend
        The I/O backend to use for accessing this partition
    """
    def __init__(
        self, meta: DataSetMeta, partition_slice: Slice,
        fileset: FileSet, start_frame: int, num_frames: int,
        io_backend: IOBackend,
        decoder: Optional[Decoder] = None,
    ):
        super().__init__(
            meta=meta,
            partition_slice=partition_slice,
            io_backend=io_backend,
            decoder=decoder,
        )
        if start_frame < self.meta.image_count:
            self._fileset = fileset.get_for_range(
                max(0, start_frame), max(0, start_frame + num_frames - 1)
            )
        self._start_frame = start_frame
        self._num_frames = num_frames
        self._corrections = CorrectionSet()
        self._worker_context: Optional["WorkerContext"] = None
        if num_frames <= 0:
            raise ValueError("invalid number of frames: %d" % num_frames)

    def get_locations(self):
        # Allow using any worker by default
        return None

    def get_max_io_size(self):
        # delegate to I/O backend by default:
        io_backend = self.get_io_backend()
        if io_backend is None:
            return None  # default value is set in Negotiator
        io_backend = io_backend.get_impl()
        return io_backend.get_max_io_size()

    def _get_read_ranges(self, tiling_scheme, roi=None):
        return self._fileset.get_read_ranges(
            start_at_frame=self._start_frame,
            stop_before_frame=min(self._start_frame + self._num_frames, self.meta.image_count),
            tiling_scheme=tiling_scheme,
            dtype=self.meta.raw_dtype,
            sync_offset=self.meta.sync_offset,
            roi=roi,
        )

    def get_io_backend(self):
        assert self._io_backend is not None
        return self._io_backend

    def set_corrections(self, corrections: Optional[CorrectionSet]):
        self._corrections = corrections

    def set_worker_context(self, worker_context: "WorkerContext"):
        self._worker_context = worker_context

    def get_tiles(self, tiling_scheme: TilingScheme, dest_dtype="float32",
            roi=None, array_backend: Optional[ArrayBackend] = None):
        """
        Return a generator over all DataTiles contained in this Partition.

        Note
        ----
        The DataSet may reuse the internal buffer of a tile, so you should
        directly process the tile and not accumulate a number of tiles and then work
        on them.

        Parameters
        ----------

        tiling_scheme
            According to this scheme the data will be tiled

        dest_dtype : numpy dtype
            convert data to this dtype when reading

        roi : numpy.ndarray
            Boolean array that matches the dataset navigation shape to limit the region to work on.
            With a ROI, we yield tiles from a "compressed" navigation axis, relative to
            the beginning of the dataset. Compressed means, only frames that have a 1
            in the ROI are considered, and the resulting tile slices are from a coordinate
            system that has the shape `(np.count_nonzero(roi),)`.

        array_backend : ArrayBackend
            Specify array backend to use. By default the first entry in the list of supported
            backends is used.

            .. versionadded:: 0.11.0
        """
        if self._start_frame < self.meta.image_count:
            dest_dtype = np.dtype(dest_dtype)
            tiling_scheme_adj = tiling_scheme.adjust_for_partition(self)
            self.validate_tiling_scheme(tiling_scheme_adj)
            read_ranges = self._get_read_ranges(tiling_scheme_adj, roi)
            io_backend = self.get_io_backend().get_impl()
            if array_backend is None:
                array_backend = self.meta.array_backends[0]

            yield from io_backend.get_tiles(
                tiling_scheme=tiling_scheme_adj, fileset=self._fileset,
                read_ranges=read_ranges, roi=roi,
                native_dtype=self.meta.raw_dtype,
                read_dtype=dest_dtype,
                sync_offset=self.meta.sync_offset,
                decoder=self._decoder,
                corrections=self._corrections,
                array_backend=array_backend,
            )

    def __repr__(self):
        return "<%s [%d:%d]>" % (
            self.__class__.__name__,
            self._start_frame, self._start_frame + self._num_frames
        )



# Functions freq_array, aperture_function, are adopted from HyperSpy
# and are subject of following copyright:
#
#  Copyright 2007-2016 The HyperSpy developers
#
#  HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with  HyperSpy.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 The LiberTEM developers
#
#  LiberTEM is distributed under the terms of the GNU General
# Public License as published by the Free Software Foundation,
# version 3 of the License.
# see: https://github.com/LiberTEM/LiberTEM

import numpy as np

from libertem.common.math import prod
from libertem.udf import UDF


def freq_array(shape, sampling=(1., 1.)):
    """
    Makes up a frequency array.

    Parameters
    ----------
    shape : (int, int)
        The shape of the array.
    sampling: (float, float), optional, (Default: (1., 1.))
        The sampling rates of the array.
    Returns
    -------
        Array of the frequencies.
    """
    f_freq_1d_y = np.fft.fftfreq(shape[0], sampling[0])
    f_freq_1d_x = np.fft.fftfreq(shape[1], sampling[1])
    f_freq_mesh = np.meshgrid(f_freq_1d_x, f_freq_1d_y)
    f_freq = np.hypot(f_freq_mesh[0], f_freq_mesh[1])

    return f_freq


def aperture_function(r, apradius, rsmooth):
    """
    A smooth aperture function that decays from apradius-rsmooth to apradius+rsmooth.
    Parameters
    ----------
    r : ndarray
        Array of input data (e.g. frequencies)
    apradius : float
        Radius (center) of the smooth aperture. Decay starts at apradius - rsmooth.
    rsmooth : float
        Smoothness in halfwidth. rsmooth = 1 will cause a decay from 1 to 0 over 2 pixel.
    Returns
    -------
        2d array containing aperture
    """

    return 0.5 * (1. - np.tanh((np.absolute(r) - apradius) / (0.5 * rsmooth)))


class HoloReconstructUDF(UDF):
    """
    Reconstruct off-axis electron holograms using a Fourier-based method.

    Running :meth:`~libertem.api.Context.run_udf` on an instance of this class
    will reconstruct a complex electron wave. Use the :code:`wave` key to access
    the raw data in the result.

    See :ref:`holography app` for detailed application example

    .. versionadded:: 0.3.0

    Parameters
    ----------

    out_shape : (int, int)
        Shape of the returned complex wave image. Note that the result should fit into the
        main memory.
        See :ref:`holography app` for more details

    sb_position : tuple, or vector
        Coordinates of sideband position with respect to non-shifted FFT of a hologram

    sb_size : float
        Radius of side band filter in pixels

    sb_smoothness : float, optional (Default: 0.05)
        Fraction of `sb_size` over which the edge of the filter aperture to be smoothed

    precision : bool, optional, (Default: True)
        Defines precision of the reconstruction, True for complex128 for the resulting
        complex wave, otherwise results will be complex64

    Examples
    --------
    >>> shape = tuple(dataset.shape.sig)
    >>> sb_position = [2, 3]
    >>> sb_size = 4.4
    >>> holo_udf = HoloReconstructUDF(out_shape=shape,
    ...                               sb_position=sb_position,
    ...                               sb_size=sb_size)
    >>> wave = ctx.run_udf(dataset=dataset, udf=holo_udf)['wave'].data
    """
    def __init__(self,
                 out_shape,
                 sb_position,
                 sb_size,
                 sb_smoothness=.05,
                 precision=True):
        if len(sb_position) != 2:
            raise ValueError(f"invalid sb_position {sb_position!r}, must be tuple of length 2")
        super().__init__(out_shape=out_shape,
                         sb_position=sb_position,
                         sb_size=sb_size,
                         sb_smoothness=sb_smoothness,
                         precision=precision)

    def get_result_buffers(self):
        """
        Initializes :class:`~libertem.common.buffers.BufferWrapper` objects for reconstructed
        wave function

        Returns
        -------
        A dictionary that maps 'wave' to the corresponding
        :class:`~libertem.common.buffers.BufferWrapper` objects
        """
        extra_shape = self.params.out_shape
        if not self.params.precision:
            dtype = np.complex64
        else:
            dtype = np.complex128
        return {
            "wave": self.buffer(kind="nav", dtype=dtype, extra_shape=extra_shape)
        }

    def get_task_data(self):
        """
        Updates `task_data`

        Returns
        -------
        kwargs : dict
        A dictionary with the following keys:
            kwargs['aperture'] : array-like
            Side band filter aperture (mask)
            kwargs['slice'] : slice
            Slice for slicing FFT of the hologram
        """

        out_shape = self.params.out_shape
        sy, sx = self.meta.partition_shape.sig
        oy, ox = out_shape
        f_sampling = (1. / oy, 1. / ox)
        sb_size = self.params.sb_size * np.mean(f_sampling)
        sb_smoothness = sb_size * self.params.sb_smoothness * np.mean(f_sampling)

        f_freq = freq_array(out_shape)
        aperture = aperture_function(f_freq, sb_size, sb_smoothness)

        y_min = int(sy / 2 - oy / 2)
        y_max = int(sy / 2 + oy / 2)
        x_min = int(sx / 2 - ox / 2)
        x_max = int(sx / 2 + oy / 2)
        slice_fft = (slice(y_min, y_max), slice(x_min, x_max))

        kwargs = {
            'aperture': self.xp.array(aperture),
            'slice': slice_fft
        }
        return kwargs

    def process_frame(self, frame):
        """
        Reconstructs holograms outputting results into 'wave'

        Parameters
        ----------
        frame
           single frame (hologram) of the data
        """
        if not self.params.precision:
            frame = frame.astype(np.float32)
        # size_x, size_y = self.params.out_shape
        frame_size = self.meta.partition_shape.sig
        sb_pos = self.params.sb_position
        aperture = self.task_data.aperture
        slice_fft = self.task_data.slice

        fft_frame = self.xp.fft.fft2(frame) / prod(frame_size)
        fft_frame = self.xp.roll(fft_frame, sb_pos, axis=(0, 1))

        fft_frame = self.xp.fft.fftshift(self.xp.fft.fftshift(fft_frame)[slice_fft])

        fft_frame = fft_frame * aperture

        wav = self.xp.fft.ifft2(fft_frame) * prod(frame_size)
        # FIXME check if result buffer with where='device' and export is faster
        # than exporting frame by frame, as implemented now.
        if self.meta.device_class == 'cuda':
            # That means xp is cupy
            wav = self.xp.asnumpy(wav)
        self.results.wave[:] = wav

    def get_backends(self):
        # CuPy support deactivated due to https://github.com/LiberTEM/LiberTEM/issues/815
        return ('numpy',)
        # return ('numpy', 'cupy')



import os
import re
import glob
import math
import typing
import logging
import itertools

import numpy as np
import numba
from numba.typed import List
from ncempy.io import dm

from libertem.common.math import prod, make_2D_square, flat_nonzero
from libertem.common.buffers import zeros_aligned
from libertem.common import Shape
from libertem.common.messageconverter import MessageConverter
from .base import (
    DataSet, BasePartition, DataSetException, DataSetMeta,
    FileSet, File, Decoder, make_get_read_ranges,
    TilingScheme, IOBackend,
)
from .base.file import OffsetsSizes

log = logging.getLogger(__name__)


HEADER_SIZE = 40
BLOCK_SIZE = 0x5758
DATA_SIZE = BLOCK_SIZE - HEADER_SIZE
BLOCK_SHAPE = (930, 16)
BLOCKS_PER_SECTOR_PER_FRAME = 32
FIRST_FRAME_SCAN_SIZE = 400
NUM_SECTORS = 8
SECTOR_SIZE = (2 * 930, 256)

SHUTTER_ACTIVE_MASK = 0x1


class K2ISDatasetParams(MessageConverter):
    SCHEMA = {
      "$schema": "http://json-schema.org/draft-07/schema#",
      "$id": "http://libertem.org/K2ISDatasetParams.schema.json",
      "title": "K2ISDatasetParams",
      "type": "object",
      "properties": {
          "type": {"const": "K2IS"},
          "path": {"type": "string"},
          "nav_shape": {
              "type": "array",
              "items": {"type": "number", "minimum": 1},
              "minItems": 2,
              "maxItems": 2
          },
          "sig_shape": {
              "type": "array",
              "items": {"type": "number", "minimum": 1},
              "minItems": 2,
              "maxItems": 2
          },
          "sync_offset": {"type": "number"},
          "io_backend": {
              "enum": IOBackend.get_supported(),
          },
      },
      "required": ["type", "path"]
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["path"]
        }
        if "nav_shape" in raw_data:
            data["nav_shape"] = tuple(raw_data["nav_shape"])
        if "sig_shape" in raw_data:
            data["sig_shape"] = tuple(raw_data["sig_shape"])
        if "sync_offset" in raw_data:
            data["sync_offset"] = raw_data["sync_offset"]
        return data


@numba.njit(inline='always')
def decode_uint12_le(inp, out):
    """
    Decode bytes from bytestring ``inp`` as 12 bit into ``out``

    Based partially on https://stackoverflow.com/a/45070947/540644
    """
    # assert np.mod(len(inp), 3) == 0
    # assert len(out) >= len(inp) * 2 / 3

    for i in range(len(inp) // 3):
        fst_uint8 = np.uint16(inp[i * 3])
        mid_uint8 = np.uint16(inp[i * 3 + 1])
        lst_uint8 = np.uint16(inp[i * 3 + 2])

        a = fst_uint8 | (mid_uint8 & 0x0F) << 8
        b = (mid_uint8 & 0xF0) >> 4 | lst_uint8 << 4
        out[i * 2] = a
        out[i * 2 + 1] = b


# @numba.njit(inline='always', boundscheck=True)
@numba.njit(inline='always')
def decode_k2is(inp, out, idx, native_dtype, rr, origin, shape, ds_shape):
    """
    Decode a single block, from a single read range, into a tile that may
    contain multiple blocks in the signal dimensions. This function is called
    multiple times for a single tile, for all read ranges that are part of this
    tile.
    """
    # blocks per tile (only in signal dimensions)
    blocks_per_tile = out.shape[1] // (BLOCK_SHAPE[0] * BLOCK_SHAPE[1])

    n_blocks_y, n_blocks_x, block_y_i, block_x_i = rr[3:]

    tile_idx = idx // blocks_per_tile

    # we take three bytes of the input to decode two numbers
    # -> as the blocks are 16 numbers wide, the two numbers
    # will always be part of the same row, so the y-stride is the same

    # the offset between two rows in the output (in indices, not bytes)
    stride_y = shape[2]

    # shortcut, in case we don't need to handle multiple
    # blocks in x direction:
    if stride_y == 16:
        block_out = out[tile_idx, 16 * 930 * block_y_i:16 * 930 * (block_y_i + 1)]
        return decode_uint12_le(inp=inp, out=block_out)

    # starting offset of the current block:
    # 930 * block_y_i:930 * (block_y_i + 1),
    # 16 * block_x_i:16 * (block_x_i + 1),
    block_offset_y = block_y_i * BLOCK_SHAPE[0] * n_blocks_x * BLOCK_SHAPE[1]
    block_offset = block_offset_y + block_x_i * BLOCK_SHAPE[1]

    out_z = out[tile_idx]

    # decode_uint12_le(inp=inp, out=blockbuf) inlined here:
    # inp is uint8, so the outer loop needs to jump 24 bytes each time.
    for i in range(len(inp) // 3 // 8):
        # i is the output row index of a single block,
        # so the beginning of the row in output coordinates:
        out_pos = block_offset + i * stride_y

        in_row_offset = i * 3 * 8

        # loop for a single row:
        # for each j, we process bytes of input into two output numbers
        # -> we consume 8*3 = 24 bytes and generate 8*2=16 numbers
        for j in range(8):
            triplet_offset = in_row_offset + j * 3
            fst_uint8 = np.uint16(inp[triplet_offset])
            mid_uint8 = np.uint16(inp[triplet_offset + 1])
            lst_uint8 = np.uint16(inp[triplet_offset + 2])

            a = fst_uint8 | (mid_uint8 & 0x0F) << 8
            b = (mid_uint8 & 0xF0) >> 4 | lst_uint8 << 4

            out_z[out_pos] = a
            out_z[out_pos + 1] = b

            out_pos += 2


class K2ISDecoder(Decoder):
    def get_decode(self, native_dtype, read_dtype):
        return decode_k2is


@numba.njit(inline='always')
def _k2is_read_ranges_tile_block(
    slices_arr, fileset_arr, slice_sig_sizes, sig_origins,
    inner_indices_start, inner_indices_stop, frame_indices, sig_size,
    px_to_bytes, bpp, frame_header_bytes, frame_footer_bytes, file_idxs,
    slice_offset, extra, sig_shape,
):
    result = List()

    # positions in the signal dimensions:
    for slice_idx in range(slices_arr.shape[0]):
        # (offset, size) arrays defining what data to read (in pixels)
        slice_origin = slices_arr[slice_idx][0]
        slice_shape = slices_arr[slice_idx][1]

        read_ranges = List()

        n_blocks_y = slice_shape[0] // 930
        n_blocks_x = slice_shape[1] // 16

        origin_block_y = slice_origin[0] // 930
        origin_block_x = slice_origin[1] // 16

        # inner "depth" loop along the (flat) navigation axis of a tile:
        for i, inner_frame_idx in enumerate(range(inner_indices_start, inner_indices_stop)):
            inner_frame = frame_indices[inner_frame_idx]
            frame_in_file_idx = inner_frame  # in k2is all files contain data from all frames

            for block_y_i in range(n_blocks_y):
                sector_index_y = origin_block_y + block_y_i
                for block_x_i in range(n_blocks_x):
                    block_index_x = origin_block_x + block_x_i
                    sector_id = block_index_x // 16
                    sector_index_x = block_index_x % 16
                    f = fileset_arr[sector_id]
                    file_header_bytes = f[3]

                    # "linear" block index per sector:
                    blockidx = (15 - sector_index_x) + sector_index_y * 16
                    offset = (
                        file_header_bytes
                        + frame_in_file_idx * BLOCK_SIZE * BLOCKS_PER_SECTOR_PER_FRAME
                        + blockidx * BLOCK_SIZE
                    )
                    start = offset + HEADER_SIZE
                    stop = offset + HEADER_SIZE + DATA_SIZE
                    read_ranges.append(
                        (sector_id, start, stop,
                         n_blocks_y, n_blocks_x,
                         block_y_i, block_x_i)
                    )

        # the indices are compressed to the selected frames
        compressed_slice = np.array([
            [slice_offset + inner_indices_start] + [i for i in slice_origin],
            [inner_indices_stop - inner_indices_start] + [i for i in slice_shape],
        ])

        result.append((slice_idx, compressed_slice, read_ranges))
    return result


k2is_get_read_ranges = make_get_read_ranges(
    read_ranges_tile_block=_k2is_read_ranges_tile_block
)


def _pattern(path):
    path, ext = os.path.splitext(path)
    ext = ext.lower()
    if ext == ".gtg":
        pattern = "%s*.bin" % glob.escape(path)
    elif ext == ".bin":
        pattern = "%s*.bin" % (
            glob.escape(re.sub(r'[0-9]+$', '', path))
        )
    else:
        raise DataSetException("unknown extension: %s" % ext)
    return pattern


def get_filenames(path, disable_glob=False):
    if disable_glob:
        return [path]
    else:
        return glob.glob(_pattern(path))


def _get_gtg_path(full_path):
    path, ext = os.path.splitext(full_path)
    ext = ext.lower()
    if ext == ".gtg":
        return full_path
    elif ext == ".bin":
        return "%s.gtg" % (
            re.sub(r'[0-9]+$', '', path)
        )


def _get_nav_shape(path):
    with dm.fileDM(_get_gtg_path(path), on_memory=True) as dm_file:
        nav_shape_y = dm_file.allTags.get('.SI Dimensions.Size Y')
        nav_shape_x = dm_file.allTags.get('.SI Dimensions.Size X')
        if nav_shape_y is not None and nav_shape_x is not None:
            return (int(nav_shape_y), int(nav_shape_x))
        return None


def _nav_shape_without_flyback(path):
    with dm.fileDM(_get_gtg_path(path), on_memory=True) as dm_file:
        nav_shape_y = dm_file.allTags.get(
            '.SI Image Tags.SI.Acquisition.Spatial Sampling.Height (pixels)'
        )
        nav_shape_x = dm_file.allTags.get(
            '.SI Image Tags.SI.Acquisition.Spatial Sampling.Width (pixels)'
        )
        if nav_shape_y is not None and nav_shape_x is not None:
            return (int(nav_shape_y), int(nav_shape_x))
        return None


def _get_num_frames_w_shutter_active_flag_set(syncer):
    syncer.sync()
    with syncer.sectors[0] as s:
        num_frames = (
            s.last_block_offset - s.first_block_offset + BLOCK_SIZE
        ) // BLOCK_SIZE // BLOCKS_PER_SECTOR_PER_FRAME
    return num_frames


def _get_num_frames(syncer):
    syncer.sync_sectors()
    with syncer.sectors[0] as s:
        num_frames = (
            s.last_block_offset - s.first_block_offset + BLOCK_SIZE
        ) // BLOCK_SIZE // BLOCKS_PER_SECTOR_PER_FRAME
    return num_frames


def _get_syncer_for_detect_params(files):
    return K2Syncer(files)


class K2Syncer:
    """
    Sync the 8 sectors of a K2IS data set. First, find the first complete frame and the
    last complete frame. Next, sync to the first frame with the shutter_active flag set.
    Finally, validate the first and last frames.

    Parameters
    ----------
    paths: list of str
        List of paths of the 8 .bin files

    start_offsets: list of int, optional
        List of first block offsets of the 8 sectors

    last_offsets: list of int, optional
        List of last block offsets of the 8 sectors
    """

    def __init__(self, paths, start_offsets=None, last_offsets=None):
        self.paths = paths
        if start_offsets is None:
            start_offsets = NUM_SECTORS * [0]
            last_offsets = NUM_SECTORS * [None]
        self.sectors = [
            Sector(fname, idx, initial_offset=start_offset, end_offset=last_offset)
            for (
                (idx, fname), start_offset, last_offset
            ) in zip(enumerate(paths), start_offsets, last_offsets)
        ]

    def sync_sectors(self):
        for b in self.first_blocks():
            assert b.is_valid, "first block is not valid!"
        # sync up all sectors to start with the same `block_count`
        block_with_max_idx = sorted(self.first_blocks(), key=lambda b: b.header['block_count'])[-1]
        start_blocks = [
            s.first_block_with(
                lambda b: b.header['block_count'] == block_with_max_idx.header['block_count']
            )
            for s in self.sectors
        ]
        for b in start_blocks:
            assert b.is_valid
            b.sector.set_first_block_offset(b.offset)
        log.debug("first_block_offsets #1: %r", [s.first_block_offset for s in self.sectors])
        # skip incomplete frames:
        # if the next 32 blocks of a sector don't have all the same frame id,
        # the frame is incomplete
        have_overlap = (
            len({
                b.header['frame_id']
                for b in itertools.islice(s.get_blocks(), BLOCKS_PER_SECTOR_PER_FRAME)
            }) > 1
            for s in self.sectors
        )
        if any(have_overlap):
            log.debug("have_overlap, finding next frame")
            frame_id = self.first_blocks()[0].header['frame_id']
            for s in self.sectors:
                offset = s.first_block_with(lambda b: b.header['frame_id'] != frame_id).offset
                s.set_first_block_offset(offset)
        log.debug("first_block_offsets #2: %r", [s.first_block_offset for s in self.sectors])
        for b in self.first_blocks():
            assert b.is_valid
        # find last_block_offsets
        for s in self.sectors:
            offset = s.first_block_offset
            while offset + BLOCK_SIZE <= s.last_block_offset:
                offset += BLOCK_SIZE
            s.set_last_block_offset(offset)
        log.debug("last_block_offsets #1: %r", [s.last_block_offset for s in self.sectors])
        for b in self.last_blocks():
            assert b.is_valid, "last block is not valid!"
        # end all sectors with the same `block_count`
        block_with_min_idx = sorted(self.last_blocks(), key=lambda b: b.header['block_count'])[0]
        end_blocks = [
            s.first_block_from_end_with(
                lambda b: b.header['block_count'] == block_with_min_idx.header['block_count']
            )
            for s in self.sectors
        ]
        for b in end_blocks:
            assert b.is_valid
            b.sector.set_last_block_offset(b.offset)
        log.debug("last_block_offsets #2: %r", [s.last_block_offset for s in self.sectors])
        # skip incomplete frames:
        # if the next 32 blocks from the end of a sector don't have all the same frame id,
        # the frame is incomplete
        have_overlap_at_end = (
            len({
                b.header['frame_id']
                for b in itertools.islice(s.get_blocks_from_end(), BLOCKS_PER_SECTOR_PER_FRAME)
            }) > 1
            for s in self.sectors
        )
        if any(have_overlap_at_end):
            log.debug("have_overlap, finding previous frame")
            frame_id = self.last_blocks()[0].header['frame_id']
            for s in self.sectors:
                offset = s.first_block_from_end_with(
                    lambda b: b.header['frame_id'] != frame_id
                ).offset
                s.set_last_block_offset(offset)
        log.debug("last_block_offsets #3: %r", [s.last_block_offset for s in self.sectors])
        for b in self.last_blocks():
            assert b.is_valid

    def sync_to_first_frame(self):
        log.debug("synchronizing to shutter_active flag...")
        for s in self.sectors:
            offset = s.first_block_with_search(lambda b: b.shutter_active).offset
            s.set_first_block_offset(offset)
        log.debug("first_block_offsets #3: %r", [s.first_block_offset for s in self.sectors])

    def validate_sync(self):
        # first blocks should be valid:
        first_blocks = self.first_blocks()
        frame_id = first_blocks[0].header['frame_id']
        for b in first_blocks:
            assert b.is_valid
            assert b.header['frame_id'] == frame_id

        # in each sector, a whole frame should follow, and frame ids should match:
        for s in self.sectors:
            blocks = itertools.islice(s.get_blocks(), BLOCKS_PER_SECTOR_PER_FRAME)
            assert all(b.header['frame_id'] == frame_id
                       for b in blocks)

        # last blocks should be valid:
        last_blocks = self.last_blocks()
        frame_id = last_blocks[0].header['frame_id']
        for b in last_blocks:
            assert b.is_valid
            assert b.header['frame_id'] == frame_id

        # each sector should end with a whole frame, and frame ids should match:
        for s in self.sectors:
            blocks = itertools.islice(s.get_blocks_from_end(), BLOCKS_PER_SECTOR_PER_FRAME)
            assert all(b.header['frame_id'] == frame_id
                       for b in blocks)

    def sync(self):
        self.sync_sectors()
        self.sync_to_first_frame()
        self.validate_sync()

    def first_blocks(self):
        return [next(s.get_blocks()) for s in self.sectors]

    def last_blocks(self):
        return [next(s.get_blocks_from_end()) for s in self.sectors]

    def close(self):
        for s in self.sectors:
            s.close()


class Sector:
    def __init__(self, fname, idx, initial_offset=0, end_offset=None):
        self.fname = fname
        self.idx = idx
        self.filesize = os.stat(fname).st_size
        self.first_block_offset = initial_offset
        if end_offset is None:
            self.last_block_offset = self.filesize + self.first_block_offset - BLOCK_SIZE
        else:
            self.last_block_offset = end_offset
        # FIXME: hardcoded sig_dims
        self.sig_dims = 2

    def open(self):
        self.f = open(self.fname, "rb")

    def close(self):
        self.f.close()
        self.f = None

    def seek(self, pos):
        self.f.seek(pos)

    def read(self, size):
        return self.f.read(size)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get_block_by_index(self, idx):
        offset = self.first_block_offset + idx * BLOCK_SIZE
        return DataBlock(offset=offset, sector=self)

    def get_blocks(self):
        offset = self.first_block_offset
        while offset <= self.last_block_offset:
            yield DataBlock(offset=offset, sector=self)
            offset += BLOCK_SIZE

    def get_blocks_from_end(self):
        offset = self.last_block_offset
        while offset >= self.first_block_offset:
            yield DataBlock(offset=offset, sector=self)
            offset -= BLOCK_SIZE

    def get_block_by_offset(self, offset):
        return DataBlock(offset=offset, sector=self)

    def get_blocks_from_offset(self, offset):
        while offset <= self.last_block_offset:
            yield DataBlock(offset=offset, sector=self)
            offset += BLOCK_SIZE

    def set_first_block_offset(self, offset):
        self.first_block_offset = offset

    def set_last_block_offset(self, offset):
        self.last_block_offset = offset

    def first_block(self):
        return next(self.get_blocks())

    def last_block(self):
        return next(self.get_blocks_from_end())

    def first_block_with(self, predicate=lambda b: True):
        return next(b for b in self.get_blocks()
                    if predicate(b))

    def first_block_from_end_with(self, predicate=lambda b: True):
        return next(b for b in self.get_blocks_from_end()
                    if predicate(b))

    def first_block_with_search(self, predicate=lambda b: True, step=32 * 8 * 50):
        """
        Binary search variant of `first_block_with`, assuming that predicate is true
        from some index on, and stays true (at least for 10 * step).
        """
        # find a rough upper bound:
        upper = None
        for idx in range(0, step * 10, step):
            block = self.get_block_by_index(idx)
            if predicate(block):
                upper = idx
                break

        # the block is somewhere in [0, upper]
        def _rec(current_lower, current_upper):
            if current_upper == current_lower:
                # we found our match, return block
                assert predicate(self.get_block_by_index(current_upper))
                return self.get_block_by_index(current_upper)
            mid = math.floor((current_lower + current_upper) / 2)
            if predicate(self.get_block_by_index(mid)):
                # mid block has predicate=True, so result must be mid or lower:
                return _rec(current_lower, mid)
            else:
                # mid block has predicate=False, so result must be mid + 1 or higher:
                return _rec(mid + 1, current_upper)

        return _rec(0, upper)

    def __repr__(self):
        return "<Sector %d>" % self.idx


class DataBlock:
    header_dtype = [
        ('sync', '>u4'),
        ('padding1', (bytes, 4)),
        ('version', '>u1'),
        ('flags', '>u1'),
        ('padding2', (bytes, 6)),
        ('block_count', '>u4'),
        ('width', '>u2'),
        ('height', '>u2'),
        ('frame_id', '>u4'),
        ('pixel_x_start', '>u2'),  # first pixel x coordinate within sector
        ('pixel_y_start', '>u2'),  # first pixel y coordinate within sector
        ('pixel_x_end', '>u2'),  # last pixel x coordinate within sector
        ('pixel_y_end', '>u2'),  # last pixel y coordinate within sector
        ('block_size', '>u4'),  # should be fixed 0x5758
    ]

    def __init__(self, offset, sector):
        self.offset = offset
        self.sector = sector
        self._header_raw = None
        self._header = None
        self._data_raw = None

    @property
    def is_valid(self):
        return (
            self.sector.filesize >= self.offset + BLOCK_SIZE
            and self.header['width'] == 256
            and self.header['height'] == 1860
            and self.header['sync'] == 0xFFFF0055
        )

    @property
    def header(self):
        if self._header_raw is None:
            with self.sector:
                self.sector.seek(self.offset)
                self._header_raw = np.fromfile(self.sector.f, dtype=self.header_dtype, count=1)
        if self._header is not None:
            return self._header
        header = {}
        for field, dtype in self.header_dtype:
            if type(dtype) is not str:
                continue
            header[field] = self._header_raw[field][0]
        self._header = header
        return header

    @property
    def shutter_active(self):
        return self.header['flags'] & SHUTTER_ACTIVE_MASK == 1

    @property
    def pixel_data_raw(self):
        if self._data_raw is None:
            with self.sector:
                self.sector.seek(self.offset + HEADER_SIZE)
                self._data_raw = self.sector.read(BLOCK_SIZE - HEADER_SIZE)
        return self._data_raw

    def readinto(self, out):
        out = out.reshape(930 * 16)
        decode_uint12_le(inp=self.pixel_data_raw, out=out)
        return out.reshape(930, 16)

    @property
    def pixel_data(self):
        if not self.is_valid:
            raise ValueError("invalid block: %r" % self)
        arr = zeros_aligned((930 * 16), dtype="uint16")
        self.readinto(arr)
        return arr.reshape(930, 16)

    def copy_to_frame(self, frame):
        sector_width = 256
        x_offset = self.sector.idx * sector_width
        h = self.header
        self.readinto(frame[
            h['pixel_y_start']:h['pixel_y_end'] + 1,
            h['pixel_x_start'] + x_offset:h['pixel_x_end'] + 1 + x_offset,
        ])

    def __repr__(self):
        h = self.header
        return "<DataBlock for frame=%d x=%d:%d y=%d:%d @ %d>" % (
            h['frame_id'],
            h['pixel_x_start'], h['pixel_x_end'],
            h['pixel_y_start'], h['pixel_y_end'],
            self.offset,
        )


class K2FileSet(FileSet):
    def get_for_range(self, start, stop):
        """
        K2 files have a different mapping than other file formats, so we need
        to override this to not filter in navigation axis
        """
        return self

    def get_read_ranges(
        self, start_at_frame: int, stop_before_frame: int,
        dtype, tiling_scheme: TilingScheme, sync_offset: int = 0,
        roi: typing.Union[np.ndarray, None] = None,
    ):
        fileset_arr = self.get_as_arr()
        roi_nonzero = None
        if roi is not None:
            roi_nonzero = flat_nonzero(roi)
        kwargs = dict(
            start_at_frame=start_at_frame,
            stop_before_frame=stop_before_frame,
            roi_nonzero=roi_nonzero,
            depth=tiling_scheme.depth,
            slices_arr=tiling_scheme.slices_array,
            fileset_arr=fileset_arr,
            sig_shape=tuple(tiling_scheme.dataset_shape.sig),
            sync_offset=sync_offset,
            bpp=np.dtype(dtype).itemsize * 8,
            frame_header_bytes=self._frame_header_bytes,
            frame_footer_bytes=self._frame_footer_bytes,
        )
        return k2is_get_read_ranges(**kwargs)


class K2ISFile(File):
    def get_offsets_sizes(self, size: int) -> OffsetsSizes:
        """
        The simple "offset/size/header/footer" method doesn't
        apply for the K2IS format, so we stub it out here:
        """
        return OffsetsSizes(
            file_offset=self._file_header,
            skip_end=0,
            frame_offset=0,
            frame_size=0,
        )

    def get_array_from_memview(self, mem: memoryview, slicing: OffsetsSizes) -> np.ndarray:
        mem = mem[slicing.file_offset:]
        return np.frombuffer(mem, dtype=self._native_dtype)


class K2ISDataSet(DataSet):
    """
    Read raw K2IS data sets. They consist of 8 .bin files and one .gtg file.
    Currently, data acquired using the STEMx unit is supported, metadata
    about the nav_shape is read from the .gtg file.

    Parameters
    ----------
    path: str
        Path to one of the files of the data set (either one of the .bin files or the .gtg file)

    nav_shape: tuple of int, optional
        A n-tuple that specifies the size of the navigation region ((y, x), but
        can also be of length 1 for example for a line scan, or length 3 for
        a data cube, for example)

    sig_shape: tuple of int, optional
        Signal/detector size (height, width)

    sync_offset: int, optional
        If positive, number of frames to skip from start
        If negative, number of blank frames to insert at start

    Examples
    --------

    >>> ds = ctx.load("k2is", path='./path_to_file.bin', ...)  # doctest: +SKIP
    """

    def __init__(self, path, nav_shape=None, sig_shape=None, sync_offset=None, io_backend=None):
        super().__init__(io_backend=io_backend)
        self._path = path
        self._start_offsets = None
        self._last_offsets = None
        self._files = None
        self._nav_shape = tuple(nav_shape) if nav_shape else nav_shape
        self._sig_shape = tuple(sig_shape) if sig_shape else sig_shape
        self._is_time_series = None
        self._sync_offset = None
        self._native_sync_offset = 0
        self._user_sync_offset = sync_offset
        self._cached_user_sync_offset = None

    def _do_initialize(self):
        self._files = self._get_files()
        self._set_skip_frames_and_nav_shape()
        if self._sig_shape is None:
            self._sig_shape = (SECTOR_SIZE[0], NUM_SECTORS * SECTOR_SIZE[1])
        elif int(prod(self._sig_shape)) != int(prod(
                    (SECTOR_SIZE[0], NUM_SECTORS * SECTOR_SIZE[1])
                )):
            raise DataSetException(
                "sig_shape must be of size: %s" % int(prod(
                    (SECTOR_SIZE[0], NUM_SECTORS * SECTOR_SIZE[1])
                ))
            )
        self._image_count = _get_num_frames(self._get_syncer(do_sync=False))
        self._set_sync_offset()
        self._get_syncer(do_sync=True)
        self._meta = DataSetMeta(
            shape=Shape(self._nav_shape + self._sig_shape, sig_dims=len(self._sig_shape)),
            raw_dtype=np.dtype("uint16"),
            sync_offset=self._sync_offset,
            image_count=self._image_count,
        )
        return self

    def initialize(self, executor):
        return executor.run_function(self._do_initialize)

    def _set_skip_frames_and_nav_shape(self):
        nav_shape = _get_nav_shape(self._path)
        if nav_shape is not None:
            # the sync flag appears to be set one frame too late, so
            # we compensate here by setting a negative _skip_frames value.
            # skip_frames is applied after synchronization.
            self._is_time_series = False
            self._skip_frames = -1
            if self._nav_shape is None:
                self._nav_shape = nav_shape
        else:
            # data set is time series and does not have any frames with
            # SHUTTER_ACTIVE_MASK not set. Hence skip_frames is not needed.
            self._is_time_series = True
            self._skip_frames = 0
            if self._nav_shape is None:
                self._nav_shape = (
                    _get_num_frames_w_shutter_active_flag_set(self._get_syncer(do_sync=False)),
                )

    def _set_sync_offset(self):
        self._num_frames_w_shutter_active_flag_set = _get_num_frames_w_shutter_active_flag_set(
            self._get_syncer(do_sync=False)
        )
        self._native_sync_offset = self._image_count - self._num_frames_w_shutter_active_flag_set
        if self._user_sync_offset is None:
            self._user_sync_offset = self._native_sync_offset
        self._nav_shape_product = int(prod(self._nav_shape))
        self._sync_offset_info = self._get_sync_offset_info()
        if not self._is_time_series:
            if self._user_sync_offset == self._native_sync_offset:
                self._sync_offset = 0
            elif self._user_sync_offset > self._native_sync_offset:
                self._sync_offset = self._user_sync_offset - self._native_sync_offset
            else:
                if self._user_sync_offset > 0:
                    self._skip_frames = self._user_sync_offset - self._native_sync_offset - 1
                    self._sync_offset = 0
                else:
                    self._skip_frames = -1 * self._native_sync_offset
                    self._sync_offset = self._user_sync_offset - 1
        else:
            self._sync_offset = self._user_sync_offset

    def _get_sync_offset_info(self):
        """
        Check sync_offset specified and returns number of frames skipped and inserted
        """
        if not -1*self._image_count < self._user_sync_offset < self._image_count:
            raise DataSetException(
                "sync_offset should be in (%s, %s), which is (-image_count, image_count)"
                % (-1*self._image_count, self._image_count)
            )
        return {
            "frames_skipped_start": max(0, self._user_sync_offset),
            "frames_ignored_end": max(
                0, self._image_count - self._nav_shape_product - self._user_sync_offset
            ),
            "frames_inserted_start": abs(min(0, self._user_sync_offset)),
            "frames_inserted_end": max(
                0, self._nav_shape_product - self._image_count + self._user_sync_offset
            )
        }

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    @classmethod
    def get_msg_converter(cls):
        return K2ISDatasetParams

    @classmethod
    def get_supported_extensions(cls):
        return {"gtg", "bin"}

    @classmethod
    def detect_params(cls, path, executor):
        """
        detect if path points to a file that is part of a k2is dataset.
        extension needs to be bin or gtg, and number of files of the dataset needs to match
        the number of sectors. no further checking is done yet. in certain cases there may be false
        positives, for example if users name their binary files .bin and have 8 of them
        in a directory.
        """
        try:
            files = executor.run_function(get_filenames, path)
            if len(files) != NUM_SECTORS:
                return False
        except DataSetException:
            return False
        s = executor.run_function(_get_syncer_for_detect_params, files)
        num_frames = executor.run_function(_get_num_frames, s)
        num_frames_w_shutter_active_flag_set = executor.run_function(
            _get_num_frames_w_shutter_active_flag_set, s
        )
        sync_offset = num_frames - num_frames_w_shutter_active_flag_set
        nav_shape = executor.run_function(_get_nav_shape, path)
        if nav_shape is None:
            nav_shape = make_2D_square((num_frames_w_shutter_active_flag_set,))
        return {
            "parameters": {
                "path": path,
                "nav_shape": nav_shape,
                "sig_shape": (SECTOR_SIZE[0], NUM_SECTORS * SECTOR_SIZE[1]),
                "sync_offset": sync_offset,
            },
            "info": {
                "image_count": num_frames,
                "native_sig_shape": (SECTOR_SIZE[0], NUM_SECTORS * SECTOR_SIZE[1]),
            }
        }

    def check_valid(self):
        try:
            syncer = self._get_syncer()
            syncer.sync()
        except Exception as e:
            raise DataSetException("failed to load dataset: %s" % e) from e
        return True

    def get_cache_key(self):
        gtg = _get_gtg_path(self._path)
        return {
            "gtg_path": gtg,
            "shape": tuple(self.shape),
            "sync_offset": self._sync_offset,
        }

    def get_diagnostics(self):
        with self._get_syncer().sectors[0] as sector:
            first_block = sector.first_block()
            last_block = sector.last_block()
        fs_nosync = self._get_syncer(do_sync=False)
        sector_nosync = fs_nosync.sectors[0]
        first_block_nosync = next(sector_nosync.get_blocks())
        last_block_nosync = next(sector_nosync.get_blocks_from_end())

        return [
            {"name": "first block offsets for all sectors",
             "value": ", ".join([str(s) for s in self._start_offsets])},

            {"name": "last block offsets for all sectors",
             "value": ", ".join([str(s) for s in self._last_offsets])},

            {"name": "number of frames before sync (from first sector)",
             "value": str(self._image_count)},

            {"name": "number of frames after sync (from first sector)",
             "value": str(self._num_frames_w_shutter_active_flag_set)},

            {"name": "first frame id before sync (from first sector)",
             "value": str(first_block_nosync.header['frame_id'])},

            {"name": "first frame id after sync (from first sector)",
             "value": str(first_block.header['frame_id'])},

            {"name": "last frame id before sync (from first sector)",
             "value": str(last_block_nosync.header['frame_id'])},

            {"name": "last frame id after sync (from first sector)",
             "value": str(last_block.header['frame_id'])},
        ]

    def _get_files(self):
        files = get_filenames(self._path)
        if len(files) != NUM_SECTORS:
            raise DataSetException("expected %d files at %s, found %d" % (
                NUM_SECTORS,
                _pattern(self._path),
                len(files)
            ))
        return list(sorted(files))

    def _cache_first_block_offsets(self, fs):
        self._start_offsets = [
            s.first_block_offset
            for s in fs.sectors
        ]
        # apply skip_frames value to the start_offsets
        self._start_offsets = [o + BLOCK_SIZE * self._skip_frames * BLOCKS_PER_SECTOR_PER_FRAME
                               for o in self._start_offsets]

    def _cache_last_block_offsets(self, fs):
        self._last_offsets = [
            s.last_block_offset
            for s in fs.sectors
        ]

    def _cache_user_sync_offset(self, user_sync_offset):
        self._cached_user_sync_offset = user_sync_offset

    def _get_syncer(self, do_sync=True):
        if not do_sync:
            return K2Syncer(self._files)
        if self._start_offsets is None or self._user_sync_offset != self._cached_user_sync_offset:
            sy = K2Syncer(self._files)
            sy.sync()
            self._cache_first_block_offsets(sy)
            self._cache_last_block_offsets(sy)
            self._cache_user_sync_offset(self._user_sync_offset)
        else:
            sy = K2Syncer(
                self._files, start_offsets=self._start_offsets, last_offsets=self._last_offsets
            )
        return sy

    def get_decoder(self) -> Decoder:
        return K2ISDecoder()

    def get_base_shape(self, roi):
        return (1, 930, 16)

    def _get_fileset(self):
        files = [
            K2ISFile(
                path=path,
                start_idx=0,
                end_idx=self._image_count,
                native_dtype=np.uint8,
                sig_shape=self._sig_shape,
                file_header=offset,
            )
            for path, offset in zip(self._files, self._start_offsets)
        ]
        return K2FileSet(files=files)

    def get_partitions(self):
        io_backend = self.get_io_backend()
        fileset = self._get_fileset()
        for part_slice, start, stop in self.get_slices():
            yield K2ISPartition(
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=io_backend,
                decoder=self.get_decoder(),
            )

    def __repr__(self):
        return "<K2ISDataSet for pattern={} nav_shape={}>".format(
            _pattern(self._path), self._nav_shape
        )


class K2ISPartition(BasePartition):
    def validate_tiling_scheme(self, tiling_scheme):
        a = len(tiling_scheme.shape) == 3
        b = tiling_scheme.shape[1] % 930 == 0
        c = tiling_scheme.shape[2] % 16 == 0
        if not (a and b and c):
            raise ValueError(
                "Invalid tiling scheme: needs to be aligned to blocksize (930, 16)"
            )



import json

import pytest
import websockets

from utils import assert_msg
from aio_utils import (
    create_connection, create_analysis, create_update_compound_analysis,
)

pytestmark = [pytest.mark.web_api]


def _get_raw_params(path):
    return {
        "dataset": {
            "params": {
                "type": "RAW",
                "path": path,
                "dtype": "float32",
                "sig_shape": [128, 128],
                "enable_direct": False,
                "nav_shape": [16, 16],
                "sync_offset": 0
            }
        }
    }


@pytest.mark.asyncio
async def test_start_server(base_url, http_client, default_token):
    """
    start the server and HTTP GET '/', which should
    return a 200 status code
    """
    url = f"{base_url}?token={default_token}"
    async with http_client.get(url) as response:
        assert response.status == 200
        print(await response.text())


@pytest.mark.asyncio
async def test_get_config(base_url, http_client, default_token):
    url = f"{base_url}/api/config/?token={default_token}"
    async with http_client.get(url) as response:
        assert response.status == 200
        config = await response.json()
        assert set(config.keys()) == {"status", "messageType", "config"}
        assert set(config['config'].keys()) == {
            "version", "revision", "localCores", "cwd",
            "separator", "resultFileFormats", "devices",
            "datasetTypes",
        }


@pytest.mark.asyncio
async def test_conn_is_disconnected(base_url, http_client, default_token):
    url = f"{base_url}/api/config/connection/?token={default_token}"
    async with http_client.get(url) as response:
        assert response.status == 200
        conn = await response.json()
        print(conn)
        assert conn['status'] == 'disconnected'
        assert conn['connection'] == {}


@pytest.mark.asyncio
async def test_conn_connect_local(base_url, http_client, default_token):
    url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'local',
            'numWorkers': 2,
        }
    }
    async with http_client.put(url, json=conn_details) as response:
        assert response.status == 200
        conn_resp = await response.json()
        assert conn_resp == {
            "status": "ok",
            "connection": {
                "type": "local",
                "numWorkers": 2,
            }
        }

    async with http_client.get(url) as response:
        assert response.status == 200
        conn = await response.json()
        assert conn == {
            "status": "ok",
            "connection": {
                "type": "local",
                "numWorkers": 2,
            }
        }


@pytest.mark.asyncio
async def test_cluster_create_error(base_url, http_client, default_token):
    url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'local',
            'numWorkers': 'foo',
        }
    }
    async with http_client.put(url, json=conn_details) as response:
        conn_resp = await response.json()
        assert response.status == 500
        assert_msg(conn_resp, 'CLUSTER_CONN_ERROR', status='error')


@pytest.mark.asyncio
async def test_cluster_connect_error(base_url, http_client, default_token):
    url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'TCP',
            'address': 'tcp://invalid',
        }
    }
    async with http_client.put(url, json=conn_details) as response:
        conn_resp = await response.json()
        assert response.status == 500
        assert_msg(conn_resp, 'CLUSTER_CONN_ERROR', status='error')


@pytest.mark.asyncio
async def test_initial_state_empty(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    conn_url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'tcp',
            'address': local_cluster_url,
        }
    }
    async with http_client.put(conn_url, json=conn_details) as response:
        assert response.status == 200

    # connect to ws endpoint:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert initial_msg['messageType'] == "INITIAL_STATE"
        assert initial_msg['status'] == "ok"
        assert initial_msg['datasets'] == []
        assert initial_msg['jobs'] == []
        assert initial_msg['analyses'] == []


@pytest.mark.asyncio
async def test_initial_state_w_existing_ds(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    conn_url = f"{base_url}/api/config/connection/?token={default_token}"
    conn_details = {
        'connection': {
            'type': 'tcp',
            'address': local_cluster_url,
        }
    }
    async with http_client.put(conn_url, json=conn_details) as response:
        assert response.status == 200

    # first connect has empty list of datasets:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert initial_msg['messageType'] == "INITIAL_STATE"
        assert initial_msg['status'] == "ok"
        assert initial_msg['datasets'] == []

    raw_path = default_raw._path

    ds_uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, ds_uuid, default_token,
    )
    ds_params = _get_raw_params(raw_path)
    async with http_client.put(ds_url, json=ds_params) as resp:
        assert resp.status == 200
        resp_json = await resp.json()
        assert_msg(resp_json, 'CREATE_DATASET')
        for k in ds_params['dataset']['params']:
            assert ds_params['dataset']['params'][k] == resp_json['details']['params'][k]

    # second connect has one dataset:
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert initial_msg['messageType'] == "INITIAL_STATE"
        assert initial_msg['status'] == "ok"
        assert len(initial_msg['datasets']) == 1
        assert initial_msg['datasets'][0]["id"] == ds_uuid
        assert initial_msg['datasets'][0]["params"] == {
            "type": "RAW",
            "path": raw_path,
            "dtype": "float32",
            "sig_shape": [128, 128],
            "enable_direct": False,
            "nav_shape": [16, 16],
            "shape": [16, 16, 128, 128],
            "sync_offset": 0
        }


@pytest.mark.asyncio
async def test_initial_state_analyses(
    default_raw, base_url, http_client, server_port, local_cluster_url, default_token,
):
    await create_connection(base_url, http_client, local_cluster_url, default_token)

    # first connect has empty list of datasets:
    ws_url = f"ws://127.0.0.1:{server_port}/api/events/?token={default_token}"
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert initial_msg['messageType'] == "INITIAL_STATE"
        assert initial_msg['status'] == "ok"
        assert initial_msg['datasets'] == []

    raw_path = default_raw._path

    ds_uuid = "ae5d23bd-1f2a-4c57-bab2-dfc59a1219f3"
    ds_url = "{}/api/datasets/{}/?token={}".format(
        base_url, ds_uuid, default_token,
    )
    ds_params = _get_raw_params(raw_path)
    async with http_client.put(ds_url, json=ds_params) as resp:
        assert resp.status == 200
        resp_json = await resp.json()
        assert_msg(resp_json, 'CREATE_DATASET')
        for k in ds_params['dataset']['params']:
            assert ds_params['dataset']['params'][k] == resp_json['details']['params'][k]

        async with websockets.connect(ws_url) as ws:
            initial_msg = json.loads(await ws.recv())
            assert initial_msg['messageType'] == "INITIAL_STATE"

            ca_uuid, ca_url = await create_update_compound_analysis(
                ws, http_client, base_url, ds_uuid, token=default_token,
            )

            analysis_uuid, analysis_url = await create_analysis(
                ws, http_client, base_url, ds_uuid, ca_uuid,
                details={
                    "analysisType": "SUM_FRAMES",
                    "parameters": {
                        "roi": {
                            "shape": "disk",
                            "r": 1,
                            "cx": 1,
                            "cy": 1,
                        }
                    }
                }, token=default_token,
            )

    # second connect has one dataset and one analysis:
    async with websockets.connect(ws_url) as ws:
        initial_msg = json.loads(await ws.recv())
        assert initial_msg['messageType'] == "INITIAL_STATE"
        assert initial_msg['status'] == "ok"
        assert len(initial_msg['datasets']) == 1
        assert len(initial_msg['analyses']) == 1
        assert initial_msg["analyses"][0]["details"]["parameters"] == {
            "roi": {
                "shape": "disk",
                "r": 1,
                "cx": 1,
                "cy": 1,
            },
        }



import time
import hashlib
import logging

logger = logging.getLogger(__name__)

import numba  # NOQA: E402
try:
    from numba.core.serialize import dumps
    from numba.core.caching import FunctionCache
    from numba.core.registry import dispatcher_registry, CPUDispatcher
    old_dispatch = True
except ImportError:
    try:
        # numba 0.54 changes the internal structure, so we have to adapt our imports:
        from numba.core.serialize import dumps
        from numba.core.caching import FunctionCache
        from numba.core.registry import CPUDispatcher
        from numba.core.target_extension import (
            dispatcher_registry, jit_registry, target_registry, Generic
        )
        old_dispatch = False
    except ImportError as e:
        CPUDispatcher = None
        dispatcher_registry = None
        logger.warning(
            "could not register custom numba dispatcher, disabling custom cache (%s)" % str(e)
        )
        logger.warning(
            "numba version %s" % str(numba.__version__)
        )


_cached_njit_reg = []


def cached_njit(*args, **kwargs):
    """
    Replacement for numba.njit with custom caching. Only supports usage
    with parameters, i.e.

    @cached_njit()
    def fn():
        ...
    """
    def wrapper(fn):
        # only register with the custom target if we manage to import the right
        # structures from numba:
        if dispatcher_registry is None:
            kwargs.update({'cache': True})
        else:
            kwargs.update({'_target': 'custom_cpu', 'cache': True})
            _cached_njit_reg.append(fn)
        return numba.njit(fn, *args, **kwargs)
    return wrapper


def hasher(x):
    return hashlib.sha256(x).hexdigest()


if dispatcher_registry is not None:
    class MyFunctionCache(FunctionCache):
        def _get_dependencies(self, cvar):
            deps = [cvar]
            if hasattr(cvar, 'py_func'):
                # TODO: does the cache key need to depend on any other
                # attributes of the Dispatcher?
                closure = cvar.py_func.__closure__
                deps = [cvar.py_func.__code__.co_code]
            elif hasattr(cvar, '__closure__'):
                closure = cvar.__closure__
                # if cvar is a function and closes over a Dispatcher, the
                # cache will be busted because of the uuid that is regenerated
                deps = [cvar.__code__.co_code]
            else:
                closure = None
            if closure is not None:
                for x in closure:
                    deps.extend(self._get_dependencies(x.cell_contents))
            return deps

        def _index_key(self, sig, codegen):
            """
            Compute index key for the given signature and codegen.
            It includes a description of the OS, target architecture and hashes of
            the bytecode for the function and, if the function has a __closure__,
            a hash of the cell_contents.
            """
            codebytes = self._py_func.__code__.co_code
            cvars = self._get_dependencies(self._py_func)
            if len(cvars) > 0:
                cvarbytes = dumps(cvars)
            else:
                cvarbytes = b''

            return (sig, "libertem-numba-cache", codegen.magic_tuple(), (hasher(codebytes),
                                                 hasher(cvarbytes),))

        def load_overload(self, *args, **kwargs):
            t0 = time.time()
            data = super().load_overload(*args, **kwargs)
            t1 = time.time()
            if data is None:
                logger.info(f"numba cache miss {self._name} {self._py_func}")
            else:
                logger.info(f"cache hit for {self._name}, load took {(t1 - t0):.3f}s")
            return data

    # if we can hack in our custom caching, do it:
    class MyCPUDispatcher(CPUDispatcher):
        def enable_caching(self):
            self._cache = MyFunctionCache(self.py_func)

    if old_dispatch:
        dispatcher_registry['custom_cpu'] = MyCPUDispatcher
    else:
        class MyCPU(Generic):
            pass

        dispatcher_registry[MyCPU] = MyCPUDispatcher
        jit_registry[MyCPU] = cached_njit  # FIXME: is this needed?
        target_registry["custom_cpu"] = MyCPU



import numpy as np


def scale(factor):
    '''
    .. versionadded:: 0.6.0
    '''
    return np.eye(2) * factor


def rotate(radians):
    '''
    .. versionadded:: 0.6.0
    '''
    # https://en.wikipedia.org/wiki/Rotation_matrix
    # y, x instead of x, y
    return np.array([
        (np.cos(radians), np.sin(radians)),
        (-np.sin(radians), np.cos(radians))
    ])


def rotate_deg(degrees):
    '''
    .. versionadded:: 0.6.0
    '''
    return rotate(np.pi/180*degrees)


def flip_y():
    '''
    .. versionadded:: 0.6.0
    '''
    return np.array([
        (-1, 0),
        (0, 1)
    ])


def flip_x():
    '''
    .. versionadded:: 0.6.0
    '''
    return np.array([
        (1, 0),
        (0, -1)
    ])


def identity():
    '''
    .. versionadded:: 0.6.0
    '''
    return np.eye(2)



import logging
import itertools
import numpy as np
import dask.array as da

from libertem.common import Shape, Slice
from libertem.io.dataset.base import (
    DataSet, DataSetMeta, BasePartition, File, FileSet, DataSetException
)
from libertem.io.dataset.base.backend_mmap import MMapFile, MMapBackend, MMapBackendImpl
from libertem.common.messageconverter import MessageConverter

log = logging.getLogger(__name__)


class DaskDatasetParams(MessageConverter):
    SCHEMA = {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$id": "http://libertem.org/DaskDatasetParams.schema.json",
        "title": "DaskDatasetParams",
        "type": "object",
        "properties": {
            "type": {"const": "DASK"},
            "sig_dims": {"type": "number", "minimum": 1},
            "preserve_dimensions": {"type": "boolean"},
            "min_size": {"type": "number", "minimum": 1},
        },
        "required": ["type"],
    }

    def convert_to_python(self, raw_data):
        data = {
            k: raw_data[k]
            for k in ["sig_dims", "preserve_dimensions", "min_size"]
            if k in raw_data
        }
        return data


class FakeDaskMMapFile(MMapFile):
    """
    Implementing the same interface as MMapFile, without filesystem backing
    """
    def open(self):
        # scheduler='threads' ensures that upstream computation for this array
        # chunk happens completely on this worker and not elsewhere
        self._arr = self.desc._array.compute(scheduler='threads')
        # need to be aware that Dask can create Fortran-ordered arrays
        # when .compute is called, which can lead to downstream issues when
        # np.frombuffer is called on self._mmap in the backend. Currently it seems
        # like np.frombuffer cannot handle Fortran ordering and throws a ValueError
        self._mmap = self._arr
        return self

    def close(self):
        del self._arr
        del self._mmap


class DaskBackend(MMapBackend):
    def get_impl(self):
        return DaskBackendImpl()


class DaskBackendImpl(MMapBackendImpl):
    FILE_CLS = FakeDaskMMapFile


class DaskDataSet(DataSet):
    """
    .. versionadded:: 0.9.0

    Wraps a Dask.array.array such that it can be processed by LiberTEM.
    Partitions are created to be aligned with the array chunking. When
    the array chunking is not compatible with LiberTEM the wrapper
    merges chunks until compatibility is achieved.

    The best-case scenario is for the original array to be chunked in
    the leftmost navigation dimension. If instead another navigation
    dimension is chunked then the user can set `preserve_dimension=False`
    to re-order the navigation shape to achieve better chunking for LiberTEM.
    If more than one navigation dimension is chunked, the class will do
    its best to merge chunks without creating partitions which are too large.

    LiberTEM requires that a partition contains only whole signal frames,
    so any signal dimension chunking is immediately merged by this class.

    This wrapper is most useful when the Dask array was created using
    lazy I/O via `dask.delayed`, or via `dask.array` operations.
    The major assumption is that the chunks in the array can each be
    individually evaluated without having to read or compute more data
    than the chunk itself contains. If this is not the case then this class
    could perform very poorly due to read amplification, or even crash the Dask
    workers.

    As the class performs rechunking using a merge-only strategy it will never
    split chunks which were present in the original array. If the array
    is originally very lightly chunked, then the corresponding LiberTEM partitions
    will be very large. In addition, overly-chunked arrays (for example one chunk per
    frame) can incurr excessive Dask task graph overheads and should be avoided
    where possible.

    Parameters
    ----------

    dask_array: dask.array.array
        A Dask array

    sig_dims: int
        Number of dimensions in dask_array.shape counting from the right
        to treat as signal dimensions

    preserve_dimensions: bool, optional
        If False, allow optimization of the dask_arry chunking by
        re-ordering the nav_shape to put the most chunked dimensions first.
        This can help when more than one nav dimension is chunked.

    min_size: float, optional
        The minimum partition size in bytes if the array chunking allows
        an order-preserving merge strategy. The default min_size is 128 MiB.

    io_backend: bool, optional
        For compatibility, accept an unused io_backend argument.

    Example
    --------

    >>> import dask.array as da
    >>>
    >>> d_arr = da.ones((4, 4, 64, 64), chunks=(2, -1, -1, -1))
    >>> ds = ctx.load('dask', dask_array=d_arr, sig_dims=2)

    Will create a dataset with 5 partitions split along the zeroth dimension.
    """
    # TODO add mechanism to re-order the dimensions of results automatically
    # if preserve_dimensions is set to False
    def __init__(self, dask_array, *, sig_dims, preserve_dimensions=True,
                 min_size=None, io_backend=None):
        super().__init__(io_backend=io_backend)
        if io_backend is not None:
            raise DataSetException("DaskDataSet currently doesn't support alternative I/O backends")

        self._check_array(dask_array, sig_dims)
        self._array = dask_array
        self._sig_dims = sig_dims
        self._sig_shape = self._array.shape[-self._sig_dims:]
        self._dtype = self._array.dtype
        self._preserve_dimension = preserve_dimensions
        self._min_size = min_size
        if self._min_size is None:
            # TODO add a method to determine a sensible partition byte-size
            self._min_size = self._default_min_size

    @property
    def array(self):
        return self._array

    def get_io_backend(self):
        return DaskBackend()

    def initialize(self, executor):
        self._array = self._adapt_chunking(self._array, self._sig_dims)
        self._nav_shape = self._array.shape[:-self._sig_dims]

        self._nav_shape_product = int(np.prod(self._nav_shape))
        self._image_count = self._nav_shape_product
        shape = Shape(self._nav_shape + self._sig_shape, sig_dims=self._sig_dims)
        self._meta = DataSetMeta(
            shape=shape,
            raw_dtype=np.dtype(self._dtype),
            sync_offset=0,
            image_count=self._nav_shape_product,
        )
        return self

    @property
    def dtype(self):
        return self._meta.raw_dtype

    @property
    def shape(self):
        return self._meta.shape

    @classmethod
    def get_msg_converter(cls):
        return DaskDatasetParams

    @property
    def _default_min_size(self):
        """
        Default minimum chunk size if not supplied at init
        """
        return 128 * (2**20)  # MB

    def _chunk_slices(self, array):
        chunks = array.chunks
        boundaries = tuple(tuple(self.chunks_to_slices(chunk_lengths)) for chunk_lengths in chunks)
        return tuple(itertools.product(*boundaries))

    def _adapt_chunking(self, array, sig_dims):
        n_dimension = array.ndim
        # Handle chunked signal dimensions by merging just in case
        sig_dim_idxs = [*range(n_dimension)[-sig_dims:]]
        if any([len(array.chunks[c]) > 1 for c in sig_dim_idxs]):
            original_n_chunks = [len(c) for c in array.chunks]
            array = array.rechunk({idx: -1 for idx in sig_dim_idxs})
            log.warning('Merging sig dim chunks as LiberTEM does not '
                        'support paritioning along the sig axes. '
                        f'Original n_blocks: {original_n_chunks}. '
                        f'New n_blocks: {[len(c) for c in array.chunks]}.')
        # Warn if there is no nav_dim chunking
        n_nav_chunks = [len(dim_chunking) for dim_chunking in array.chunks[:-sig_dims]]
        if set(n_nav_chunks) == {1}:
            log.warning('Dask array is not chunked in navigation dimensions, '
                        'cannot split into nav-partitions without loading the '
                        'whole dataset on each worker. '
                        f'Array shape: {array.shape}. '
                        f'Chunking: {array.chunks}. '
                        f'array size {array.nbytes / 1e6} MiB.')
            # If we are here there is nothing else to do.
            return array
        # Orient the nav dimensions so that the zeroth dimension is
        # the most chunked, this obviously changes the dataset nav_shape !
        if not self._preserve_dimension:
            n_nav_chunks = [len(dim_chunking) for dim_chunking in array.chunks[:-sig_dims]]
            nav_sort_order = np.argsort(n_nav_chunks)[::-1].tolist()
            sort_order = nav_sort_order + sig_dim_idxs
            if not np.equal(sort_order, np.arange(n_dimension)).all():
                original_shape = array.shape
                original_n_chunks = [len(c) for c in array.chunks]
                array = da.transpose(array, axes=sort_order)
                log.warning('Re-ordered nav_dimensions to improve partitioning, '
                            'create the dataset with preserve_dimensions=True '
                            'to suppress this behaviour. '
                            f'Original shape: {original_shape} with '
                            f'n_blocks: {original_n_chunks}. '
                            f'New shape: {array.shape} with '
                            f'n_blocks: {[len(c) for c in array.chunks]}.')
        # Handle chunked nav_dimensions
        # We can allow nav_dimensions to be fully chunked (one chunk per element)
        # up-to-but-not-including the first non-fully chunked dimension. After this point
        # we must merge/rechunk all subsequent nav dimensions to ensure continuity
        # of frame indexes in a flattened nav dimension. This should be removed
        # when if we allow non-contiguous flat_idx Partitions
        nav_rechunk_dict = {}
        for dim_idx, dim_chunking in enumerate(array.chunks[:-sig_dims]):
            if set(dim_chunking) == {1}:
                continue
            else:
                merge_dimensions = [*range(dim_idx + 1, n_dimension - sig_dims)]
                for merge_i in merge_dimensions:
                    if len(array.chunks[merge_i]) > 1:
                        nav_rechunk_dict[merge_i] = -1
        if nav_rechunk_dict:
            original_n_chunks = [len(c) for c in array.chunks]
            array = array.rechunk(nav_rechunk_dict)
            log.warning('Merging nav dimension chunks according to scheme '
                        f'{nav_rechunk_dict} as we cannot maintain continuity '
                        'of frame indexing in the flattened navigation dimension. '
                        f'Original n_blocks: {original_n_chunks}. '
                        f'New n_blocks: {[len(c) for c in array.chunks]}.')
        # Merge remaining chunks maintaining C-ordering until we reach a target chunk sizes
        # or a minmum number of partitions corresponding to the number of workers
        new_chunking, min_size, max_size = merge_until_target(array, self._min_size)
        if new_chunking != array.chunks:
            original_n_chunks = [len(c) for c in array.chunks]
            chunksizes = get_chunksizes(array)
            orig_min, orig_max = chunksizes.min(), chunksizes.max()
            array = array.rechunk(new_chunking)
            log.warning('Applying re-chunking to increase minimum partition size. '
                        f'n_blocks: {original_n_chunks} => {[len(c) for c in array.chunks]}. '
                        f'Min chunk size {orig_min / 1e6:.1f} => {min_size / 1e6:.1f} MiB , '
                        f'Max chunk size {orig_max / 1e6:.1f} => {max_size / 1e6:.1f} MiB.')
        return array

    def _check_array(self, array, sig_dims):
        if not isinstance(array, da.Array):
            raise DataSetException('Expected a Dask array as input, recieved '
                                   f'{type(array)}.')
        if not isinstance(sig_dims, int) and sig_dims >= 0:
            raise DataSetException('Expected non-negative integer sig_dims,'
                                   f'recieved {sig_dims}.')
        if any([np.isnan(c).any() for c in array.shape])\
           or any([np.isnan(c).any() for c in array.chunks]):
            raise DataSetException('Dask array has an unknown shape or chunk sizes '
                                   'so cannot be interpreted as a LiberTEM partitions. '
                                   'Run array.compute_compute_chunk_sizes() '
                                   'before passing to DaskDataSet, though this '
                                   'may be performance-intensive. Chunking: '
                                   f'{array.chunks}, Shape {array.shape}')
        if sig_dims >= array.ndim:
            raise DataSetException(f'Number of sig_dims {sig_dims} not compatible '
                                   f'with number of array dims {array.ndim}, '
                                   'must be able to create partitions along nav '
                                   'dimensions.')
        return True

    def check_valid(self):
        return self._check_array(self._array, self._sig_dims)

    def get_num_partitions(self):
        return len([*itertools.product(*self._array.chunks)])

    @staticmethod
    def chunks_to_slices(chunk_lengths):
        prior = 0
        for c in chunk_lengths:
            newc = c + prior
            yield slice(prior, newc)
            prior = newc

    @staticmethod
    def slices_to_shape(slices):
        return tuple(s.stop - s.start for s in slices)

    @staticmethod
    def slices_to_origin(slices):
        return tuple(s.start for s in slices)

    @staticmethod
    def flatten_nav(slices, nav_shape, sig_dims):
        """
        Because LiberTEM partitions are set up with a flat nav dimension
        we must flatten the Dask array slices. This is ensured to be possible
        by earlier calls to _adapt_chunking but should be removed if ever
        partitions are able to have >1D navigation axes.
        """
        nav_slices = slices[:-sig_dims]
        sig_slices = slices[-sig_dims:]
        start_frame = np.ravel_multi_index([s.start for s in nav_slices], nav_shape)
        end_frame = 1 + np.ravel_multi_index([s.stop - 1 for s in nav_slices], nav_shape)
        nav_slice = slice(start_frame, end_frame)
        return (nav_slice,) + sig_slices, start_frame, end_frame

    def get_slices(self):
        """
        Generates the LiberTEM slices which correspond to the chunks
        in the Dask array backing the dataset

        Generates both the flat_nav slice for creating the LiberTEM partition
        and also the full_slices used to index into the dask array
        """
        chunk_slices = self._chunk_slices(self._array)

        for full_slices in chunk_slices:
            flat_slices, start_frame, end_frame = self.flatten_nav(full_slices, self._nav_shape,
                                                                   self._sig_dims)
            flat_slice = Slice(origin=self.slices_to_origin(flat_slices),
                               shape=Shape(self.slices_to_shape(flat_slices),
                                           sig_dims=self._sig_dims))
            yield full_slices, flat_slice, start_frame, end_frame

    def _get_fileset(self):
        """
        The fileset is set up to have one 'file' per partition
        which corresponds to one 'file' per Dask chunk
        """
        partitions = []
        for full_slices, _, start, stop in self.get_slices():
            partitions.append(DaskFile(
                array_chunk=self._array[full_slices],
                path=None,
                start_idx=start,
                end_idx=stop,
                native_dtype=self._dtype,
                sig_shape=self.shape.sig
            ))
        return DaskFileSet(partitions)

    def get_partitions(self):
        """
        Partitions contain a reference to the whole array and the whole
        fileset, but the part_slice and start_frame/num_frames provided mean
        that the subsequent call to get_read_ranges() means only one 'file'
        is read/.compute(), and this corresponds to the partition *exactly*
        """
        fileset = self._get_fileset()
        for _, part_slice, start, stop in self.get_slices():
            yield DaskPartition(
                self._array,
                meta=self._meta,
                fileset=fileset,
                partition_slice=part_slice,
                start_frame=start,
                num_frames=stop - start,
                io_backend=self.get_io_backend(),
                decoder=self.get_decoder()
            )

    def __repr__(self):
        return (f"<DaskDataSet of {self.dtype} shape={self.shape}, "
                f"n_blocks={[len(c) for c in self._array.chunks]}>")


class DaskFile(File):
    def __init__(self, *args, array_chunk=None, **kwargs):
        """
        Upon creation, the dask array has been sliced to give
        only one chunk corresponding to a LiberTEM partition
        """
        self._array = array_chunk
        super().__init__(*args, **kwargs)


class DaskFileSet(FileSet):
    pass


class DaskPartition(BasePartition):
    def __init__(self, dask_array, *args, **kwargs):
        self._array = dask_array
        super().__init__(*args, **kwargs)


def array_mult(*arrays, dtype=np.float64):
    num_arrays = len(arrays)
    if num_arrays == 1:
        return np.asarray(arrays[0]).astype(dtype)
    elif num_arrays == 2:
        return np.multiply.outer(*arrays).astype(dtype)
    elif num_arrays > 2:
        return np.multiply.outer(arrays[0], array_mult(*arrays[1:]))
    else:
        raise RuntimeError('Unexpected number of arrays')


def get_last_chunked_dim(chunking):
    n_chunks = [len(c) for c in chunking]
    chunked_dims = [idx for idx, el in enumerate(n_chunks) if el > 1]
    try:
        return chunked_dims[-1]
    except IndexError:
        return -1


def get_chunksizes(array, chunking=None):
    if chunking is None:
        chunking = array.chunks
    shape = array.shape
    el_bytes = array.dtype.itemsize
    last_chunked = get_last_chunked_dim(chunking)
    if last_chunked < 0:
        return np.asarray(array.nbytes)
    static_size = np.prod(shape[last_chunked + 1:], dtype=np.float64) * el_bytes
    chunksizes = array_mult(*chunking[:last_chunked + 1]) * static_size
    return chunksizes


def modify_chunking(chunking, dim, merge_idxs):
    chunk_dim = chunking[dim]
    merge_idxs = tuple(sorted(merge_idxs))
    before = chunk_dim[:merge_idxs[0]]
    after = chunk_dim[merge_idxs[1] + 1:]
    merged_dim = (sum(chunk_dim[merge_idxs[0]:merge_idxs[1] + 1]),)
    new_chunk_dim = tuple(before) + merged_dim + tuple(after)
    chunking = chunking[:dim] + (new_chunk_dim,) + chunking[dim + 1:]
    return chunking


def findall(sequence, val):
    return [idx for idx, e in enumerate(sequence) if e == val]


def neighbour_idxs(sequence, idx):
    max_idx = len(sequence) - 1
    if idx > 0 and idx < max_idx:
        return (idx - 1, idx + 1)
    elif idx == 0:
        return (None, idx + 1)
    elif idx == max_idx:
        return (idx - 1, None)
    else:
        raise


def min_neighbour(sequence, idx):
    left, right = neighbour_idxs(sequence, idx)
    if left is None:
        return right
    elif right is None:
        return left
    else:
        return min([left, right], key=lambda x: sequence[x])


def min_with_min_neighbor(sequence):
    min_val = min(sequence)
    occurences = findall(sequence, min_val)
    min_idx_pairs = [(idx, min_neighbour(sequence, idx)) for idx in occurences]
    pair = [sum(get_values(sequence, idxs)) for idxs in min_idx_pairs]
    min_pair = min(pair)
    min_pair_occurences = findall(pair, min_pair)
    return min_idx_pairs[min_pair_occurences[-1]]  # breaking ties from right


def get_values(sequence, idxs):
    return [sequence[idx] for idx in idxs]


def merge_until_target(array, target, min_chunks=0):
    chunking = array.chunks
    if array.nbytes < target:
        # A really small dataset, better to treat as one partition
        chunking = tuple((s,) for s in array.shape)
    chunksizes = get_chunksizes(array)
    while chunksizes.size > min_chunks and chunksizes.min() < target:
        if (chunksizes < 0).any():
            log.warn('Overflow in chunksize calculation, will be clipped!')
        chunksizes = np.clip(chunksizes, 0., np.inf)
        last_chunked_dim = get_last_chunked_dim(chunking)
        if last_chunked_dim < 0:
            # No chunking, by definition complete
            break
        last_chunking = chunking[last_chunked_dim]
        to_merge = min_with_min_neighbor(last_chunking)
        chunking = modify_chunking(chunking, last_chunked_dim, to_merge)
        chunksizes = get_chunksizes(array, chunking=chunking)
    return chunking, chunksizes.min(), chunksizes.max()



import os

import pytest
import numpy as np

from libertem.common.slice import Slice, Shape
from libertem.io.dataset.base import DataTile
from libertem.io.writers.base import WriteHandle


def test_write_handle(tmpdir_factory):
    """
    test the common "happy path":
    """
    datadir = tmpdir_factory.mktemp('write_handle_tests')
    full_path = os.path.join(datadir, "f1")
    part_slice = Slice(
        shape=Shape((32, 64, 64), sig_dims=2),
        origin=(16, 0, 0),
    )
    tile_slice = Slice(
        shape=Shape((3, 64, 64), sig_dims=2),
        origin=(19, 0, 0),
    )
    tile_data = np.random.randn(3, 64, 64).astype("float32")
    tile = DataTile(
        tile_data,
        tile_slice=tile_slice,
        scheme_idx=0,
    )

    wh = WriteHandle(full_path, datadir, part_slice, dtype='float32')

    tmp_file_name = ""

    with wh:
        wh.write_tile(tile)
        tmp_file_name = wh._tmp_file.name
        assert os.path.exists(tmp_file_name)

    # check some internals:
    assert wh._dest is None
    assert wh._tmp_file is None

    # the temporary file should no longer exist in case of success
    assert not os.path.exists(tmp_file_name)

    # ... buf our dest fname should:
    assert os.path.exists(full_path)
    assert os.path.isfile(full_path)

    # check if data is written correctly:
    read_data = np.fromfile(full_path, dtype="float32").reshape(part_slice.shape)
    assert np.allclose(
        read_data[3:6, ...],
        tile_data
    )


def test_write_handle_aborted(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('write_handle_tests')
    full_path = os.path.join(datadir, "f1")
    part_slice = Slice(
        shape=Shape((32, 64, 64), sig_dims=2),
        origin=(16, 0, 0),
    )
    tile_slice = Slice(
        shape=Shape((3, 64, 64), sig_dims=2),
        origin=(19, 0, 0),
    )
    tile_data = np.random.randn(3, 64, 64).astype("float32")
    tile = DataTile(
        tile_data,
        tile_slice=tile_slice,
        scheme_idx=0,
    )

    wh = WriteHandle(full_path, datadir, part_slice, dtype='float32')

    tmp_file_name = ""

    with wh:
        wh.write_tile(tile)
        tmp_file_name = wh._tmp_file.name
        assert os.path.exists(tmp_file_name)
        wh.abort()
        assert not os.path.exists(tmp_file_name)

    # check some internals:
    assert wh._dest is None
    assert wh._tmp_file is None

    # the temporary file should no longer exist in case of abortion
    assert not os.path.exists(tmp_file_name)

    # and neither should the full destination path
    assert not os.path.exists(full_path)


def test_write_handle_exception(tmpdir_factory):
    datadir = tmpdir_factory.mktemp('write_handle_tests')
    full_path = os.path.join(datadir, "f1")
    part_slice = Slice(
        shape=Shape((32, 64, 64), sig_dims=2),
        origin=(16, 0, 0),
    )
    tile_slice = Slice(
        shape=Shape((3, 64, 64), sig_dims=2),
        origin=(19, 0, 0),
    )
    tile_data = np.random.randn(3, 64, 64).astype("float32")
    tile = DataTile(
        tile_data,
        tile_slice=tile_slice,
        scheme_idx=0,
    )

    wh = WriteHandle(full_path, datadir, part_slice, dtype='float32')

    tmp_file_name = ""

    with pytest.raises(Exception):
        with wh:
            wh.write_tile(tile)
            tmp_file_name = wh._tmp_file.name
            assert os.path.exists(tmp_file_name)
            raise Exception("nope")

    # check some internals:
    assert wh._dest is None
    assert wh._tmp_file is None

    # the temporary file should no longer exist in case of exception
    assert not os.path.exists(tmp_file_name)

    # and neither should the full destination path
    assert not os.path.exists(full_path)



import numpy as np
import pytest
from numpy.testing import assert_allclose
import random

import libertem.udf.com as com
from sparseconverter import (
    SPARSE_BACKENDS, NUMPY, SPARSE_COO, get_device_class,
    CUPY_SCIPY_CSC
)

from libertem.io.dataset.memory import MemoryDataSet

from utils import _mk_random, set_device_class


@pytest.mark.parametrize(
    'backend', (None, ) + tuple(com.CoMUDF().get_backends())
)
def test_com(lt_ctx, delayed_ctx, backend):
    with set_device_class(get_device_class(backend)):
        if backend in SPARSE_BACKENDS:
            data = _mk_random((16, 8, 32, 64), array_backend=SPARSE_COO)
        else:
            data = _mk_random((16, 8, 32, 64), array_backend=NUMPY)

        dataset = MemoryDataSet(
            data=data, tileshape=(8, 32, 64),
            num_partitions=2,
            sig_dims=2,
            array_backends=(backend, ) if backend is not None else None
        )

        com_result = lt_ctx.run_udf(udf=com.CoMUDF(), dataset=dataset)
        com_delayed = delayed_ctx.run_udf(udf=com.CoMUDF(), dataset=dataset)

        com_analysis = lt_ctx.create_com_analysis(dataset=dataset)
        analysis_result = lt_ctx.run(com_analysis)

        if backend is CUPY_SCIPY_CSC:
            pytest.xfail(
                'For unknown reasons the dot product seems to be unstable '
                'for cupyx.scipy.csc_matrix'
            )

        assert_allclose(
            com_result['field'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1),
        )
        assert_allclose(
            com_result['field_y'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 0],
        )
        assert_allclose(
            com_result['field_x'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 1],
        )
        assert_allclose(
            com_result['field'],
            np.stack((com_result['field_y'], com_result['field_x']), axis=-1),
        )
        assert_allclose(
            com_result['magnitude'],
            analysis_result.magnitude.raw_data,
        )
        assert_allclose(
            com_result['divergence'],
            analysis_result.divergence.raw_data,
        )
        assert_allclose(
            com_result['curl'],
            analysis_result.curl.raw_data,
        )

        assert_allclose(
            com_delayed['field'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1),
        )
        assert_allclose(
            com_delayed['field_y'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 0],
        )
        assert_allclose(
            com_delayed['field_x'],
            np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 1],
        )
        assert_allclose(
            com_delayed['field'],
            np.stack((com_delayed['field_y'], com_delayed['field_x']), axis=-1),
        )
        assert_allclose(
            com_delayed['magnitude'],
            analysis_result.magnitude.raw_data,
        )
        assert_allclose(
            com_delayed['divergence'],
            analysis_result.divergence.raw_data,
        )
        assert_allclose(
            com_delayed['curl'],
            analysis_result.curl.raw_data,
        )


@pytest.mark.parametrize('repeat', range(10))
def test_com_params(lt_ctx, repeat):
    data = _mk_random((16, 8, 32, 64))
    dataset = MemoryDataSet(data=data)

    cy = random.choice((None, 2.1, 17))
    cx = random.choice((None, 3.7, 23))
    ro = random.choice((float('inf'), 10.2, 19))
    ri = random.choice((0., 3.1, 9))
    if np.isinf(ro):
        # Non-zero ri not supported by analysis API
        ri = 0.
    scan_rotation = random.choice((0., 13.3, 233))
    flip_y = random.choice((True, False))

    com_udf = com.CoMUDF.with_params(
        cy=cy,
        cx=cx,
        r=ro,
        ri=ri,
        scan_rotation=scan_rotation,
        flip_y=flip_y,
    )
    com_params = com_udf.params.com_params

    com_dict = {}
    com_dict['cy'] = com_params.cy
    com_dict['cx'] = com_params.cx
    com_dict['scan_rotation'] = com_params.scan_rotation
    com_dict['flip_y'] = com_params.flip_y
    com_dict['mask_radius'] = com_params.r
    com_dict['mask_radius_inner'] = com_params.ri

    # Test the analysis default None parameters
    if np.isinf(com_dict['mask_radius']) and random.choice((True, False)):
        com_dict['mask_radius'] = None

    if (com_dict['mask_radius'] is None
       or (np.isclose(com_dict['mask_radius_inner'], 0.) and random.choice((True, False)))):
        com_dict['mask_radius_inner'] = None

    com_result = lt_ctx.run_udf(udf=com.CoMUDF(com_params), dataset=dataset)
    com_analysis = lt_ctx.create_com_analysis(dataset=dataset, **com_dict)
    analysis_result = lt_ctx.run(com_analysis)

    assert_allclose(
        com_result['field'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1),
    )
    assert_allclose(
        com_result['field_y'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 0],
    )
    assert_allclose(
        com_result['field_x'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 1],
    )
    assert_allclose(
        com_result['field'],
        np.stack((com_result['field_y'], com_result['field_x']), axis=-1),
    )
    assert_allclose(
        com_result['magnitude'],
        analysis_result.magnitude.raw_data,
    )
    assert_allclose(
        com_result['divergence'],
        analysis_result.divergence.raw_data,
    )
    assert_allclose(
        com_result['curl'],
        analysis_result.curl.raw_data,
    )


def test_com_invalid_annulus_params():
    with pytest.raises(ValueError):
        com.CoMUDF.with_params(
            r=5.,
            ri=10.,
        )


def test_invalid_sig_shape(lt_ctx):
    ds = lt_ctx.load(
        'memory',
        data=np.zeros((2, 2, 4, 4, 4)),
        sig_dims=3,
        num_partitions=1,  # avoid a warning
    )
    udf = com.CoMUDF()
    with pytest.raises(ValueError):
        lt_ctx.run_udf(ds, udf)


@pytest.mark.parametrize('dims', (1, 3))
def test_invalid_nav_shape(lt_ctx, dims):
    nav_shape = (2,) * dims
    ds = lt_ctx.load(
        'memory',
        data=np.zeros(nav_shape + (4, 4)),
        sig_dims=2,
        num_partitions=1,  # avoid a warning
    )
    udf = com.CoMUDF()
    with pytest.raises(ValueError):
        lt_ctx.run_udf(ds, udf)


@pytest.mark.parametrize('repeat', range(10))
def test_com_roi(lt_ctx, repeat):

    data = _mk_random((16, 8, 32, 64))
    dataset = MemoryDataSet(data=data)

    roi = np.random.choice([True, False], dataset.shape.nav)

    com_result = lt_ctx.run_udf(udf=com.CoMUDF(), dataset=dataset, roi=roi)

    com_analysis = lt_ctx.create_com_analysis(dataset=dataset)
    analysis_result = lt_ctx.run(com_analysis, roi=roi)

    assert_allclose(
        com_result['field'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1),
    )
    assert_allclose(
        com_result['field_y'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 0],
    )
    assert_allclose(
        com_result['field_x'],
        np.flip(np.moveaxis(analysis_result.field.raw_data, 0, -1), axis=-1)[..., 1],
    )
    assert_allclose(
        com_result['field'],
        np.stack((com_result['field_y'], com_result['field_x']), axis=-1),
    )
    assert_allclose(
        com_result['magnitude'],
        analysis_result.magnitude.raw_data,
    )

    # The Analysis may return non-NaN values outside of the ROI
    # if all neighbors are within the ROI because of the way how np.gradient works.
    # It can directly return the result for the entire nav buffer, i.e. also set
    # values outside of the ROI.
    # The UDF cannot set values outside of the ROI to exactly replicate the Analysis
    # because of the way how get_results() and UDF result buffers for kind='nav' work.
    # Instead of equality for the entire result, we test for equality within the ROI
    # and only NaN outside of the ROI for the UDF result.
    inverted_roi = np.invert(roi)
    assert_allclose(
        com_result['divergence'].data[roi],
        analysis_result.divergence.raw_data[roi],
    )
    assert np.all(np.isnan(com_result['divergence'].data[inverted_roi]))
    assert_allclose(
        com_result['curl'].data[roi],
        analysis_result.curl.raw_data[roi],
    )
    assert np.all(np.isnan(com_result['curl'].data[inverted_roi]))


@pytest.mark.parametrize(
    'use_roi', (True, False)
)
@pytest.mark.parametrize(
    'regression', (-1, 0, 1, [[0, 0], [0, 0], [0, 0]])
)
def test_com_regression_neutral(lt_ctx, use_roi, regression):
    data = np.zeros((23, 42, 23, 42))
    for yy in range(23):
        for xx in range(42):
            data[yy, xx, 3, 3] = 1
    # Make sure the first partition extends in nav y direction,
    # so num_partitions smaller than y nav shape
    ds = lt_ctx.load('memory', data=data, num_partitions=17)
    if use_roi:
        roi = np.random.choice([True, False], ds.shape.nav)
    else:
        roi = None

    udf = com.CoMUDF.with_params(cy=3, cx=3, regression=regression)
    for iter_res in lt_ctx.run_udf_iter(dataset=ds, udf=udf, roi=roi):
        res = iter_res.buffers[0]
        assert_allclose(res['regression'], 0, atol=1e-13)
        assert_allclose(res['field'].raw_data, 0, atol=1e-13)
        assert_allclose(res['field_y'].raw_data, 0, atol=1e-13)
        assert_allclose(res['field_x'].raw_data, 0, atol=1e-13)


@pytest.mark.parametrize(
    'use_roi', (True, False)
)
@pytest.mark.parametrize(
    'regression', (0, 1, [[-2, 1], [0, 0], [0, 0]])
)
def test_com_regression_offset(lt_ctx, use_roi, regression):
    data = np.zeros((23, 42, 23, 42))
    for yy in range(23):
        for xx in range(42):
            data[yy, xx, 1, 4] = 1
    ds = lt_ctx.load('memory', data=data, num_partitions=17)
    if use_roi:
        roi = np.random.choice([True, False], ds.shape.nav)
    else:
        roi = None

    udf = com.CoMUDF.with_params(cy=3, cx=3, regression=regression)
    res = lt_ctx.run_udf(dataset=ds, udf=udf, roi=roi)
    for iter_res in lt_ctx.run_udf_iter(dataset=ds, udf=udf, roi=roi):
        res = iter_res.buffers[0]
        assert_allclose(res['regression'].data, [
            [-2, 1],  # constant offset in y and x direction
            [0, 0],
            [0, 0]
        ], atol=1e-13)
        assert_allclose(res['field'].raw_data, 0, atol=1e-13)
        assert_allclose(res['field_y'].raw_data, 0, atol=1e-13)
        assert_allclose(res['field_x'].raw_data, 0, atol=1e-13)


@pytest.mark.parametrize(
    'use_roi', (True, False)
)
@pytest.mark.parametrize(
    'regression', (1, [[-2, 42], [2, 0], [0, -1]])
)
def test_com_regression_linear(lt_ctx, use_roi, regression):
    data = np.zeros((23, 42, 48, 46))
    for yy in range(23):
        for xx in range(42):
            data[yy, xx, 2*yy+1, 45-xx] = 1
    ds = lt_ctx.load('memory', data=data, num_partitions=7)
    if use_roi:
        roi = np.random.choice([True, False], ds.shape.nav)
    else:
        roi = None

    udf = com.CoMUDF.with_params(cy=3, cx=3, regression=regression)
    for iter_res in lt_ctx.run_udf_iter(dataset=ds, udf=udf, roi=roi):
        res = iter_res.buffers[0]
        damage = iter_res.damage
        assert_allclose(res['regression'].data, [[-2, 42], [2, 0], [0, -1]], atol=1e-13)
        # The linear gradient is corrected in valid data
        assert_allclose(res['field'].data[damage], 0, atol=1e-12)
        # Regression only applied to valid data, other parts remain at their
        # default value, 0 in this case
        assert_allclose(res['field'].raw_data[~damage.raw_data], 0, atol=1e-12)
        assert_allclose(res['field_y'].data[damage], 0, atol=1e-12)
        assert_allclose(res['field_x'].data[damage], 0, atol=1e-12)


@pytest.mark.parametrize(
    'use_roi', (True, False)
)
@pytest.mark.parametrize(
    'regression', (1, [[-2, 1], [2, 1], [3, 4]])
)
def test_com_regression_linear_2(lt_ctx, use_roi, regression):
    data = np.zeros((5, 8, 64, 64))
    for yy in range(5):
        for xx in range(8):
            # dy/dy: 2, dx/dy: 1
            # dy/dx: 3, dx/dx: 4
            data[yy, xx, 1+2*yy+3*xx, 4+yy+4*xx] = 1
    ds = lt_ctx.load('memory', data=data, num_partitions=3)
    if use_roi:
        roi = np.ones(ds.shape.nav, dtype=bool)
        roi[2:4, 5:7] = False
    else:
        roi = None

    udf = com.CoMUDF.with_params(cy=3, cx=3, regression=regression)
    for iter_res in lt_ctx.run_udf_iter(dataset=ds, udf=udf, roi=roi):
        res = iter_res.buffers[0]
        assert_allclose(res['regression'].data, [[-2, 1], [2, 1], [3, 4]], atol=1e-13)
        assert_allclose(res['field'].raw_data, 0, atol=1e-12)
        assert_allclose(res['field_y'].raw_data, 0, atol=1e-12)
        assert_allclose(res['field_x'].raw_data, 0, atol=1e-12)


def test_invalid_regression_val(lt_ctx, npy_8x8x8x8_ds):
    udf = com.CoMUDF.with_params(regression=5)
    with pytest.raises(ValueError):
        lt_ctx.run_udf(npy_8x8x8x8_ds, udf)
